from collections import ChainMap, defaultdict
from dataclasses import dataclass, field
from fractions import Fraction
from io import StringIO
from itertools import zip_longest
from math import ceil
from typing import Dict, Iterator, List, Union

from more_itertools import chunked, collapse, intersperse, mark_ends, windowed

from jubeatools.info import __repository_url__, __version__
from jubeatools.song import (
    BeatsTime,
    Chart,
    LongNote,
    Metadata,
    NotePosition,
    TapNote,
    Timing,
)

from ..dump_tools import (
    DEFAULT_EXTRA_SYMBOLS,
    DIRECTION_TO_ARROW,
    DIRECTION_TO_LINE,
    NOTE_TO_CIRCLE_FREE_SYMBOL,
    JubeatAnalyserDumpedSection,
    LongNoteEnd,
    create_sections_from_chart,
    make_full_dumper_from_jubeat_analyser_chart_dumper,
)
from ..symbols import NOTE_SYMBOLS

AnyNote = Union[TapNote, LongNote, LongNoteEnd]

EMPTY_BEAT_SYMBOL = "−"  # U+02212 : MINUS SIGN
EMPTY_POSITION_SYMBOL = "□"  # U+025A1 : WHITE SQUARE


@dataclass
class Frame:
    positions: Dict[NotePosition, str] = field(default_factory=dict)
    bars: Dict[int, Dict[int, str]] = field(default_factory=dict)

    def dump(self, length: BeatsTime) -> Iterator[str]:
        self.raise_if_unfit()
        for pos, bar in zip_longest(self.dump_positions(), self.dump_bars(length)):
            if bar is None:
                bar = ""
            yield f"{pos} {bar}"

    def raise_if_unfit(self) -> None:
        if not self.bars_are_contiguous():
            raise ValueError("Frame has discontinuous bars")
        if not self.all_bars_are_from_the_same_group():
            raise ValueError("Frame contains bars from different 4-bar groups")

    def bars_are_contiguous(self) -> bool:
        return all(
            b - a == 1
            for a, b in windowed(sorted(self.bars), 2)
            if b is not None and a is not None
        )

    def all_bars_are_from_the_same_group(self) -> bool:
        return self.bars.keys() == set(bar % 4 for bar in self.bars)

    def dump_positions(self) -> Iterator[str]:
        for y in range(4):
            yield "".join(
                self.positions.get(NotePosition(x, y), EMPTY_POSITION_SYMBOL)
                for x in range(4)
            )

    def dump_bars(self, length: BeatsTime) -> Iterator[str]:
        all_bars = []
        for i in range(ceil(length * 4)):
            bar_index = i // 4
            time_index = i % 4
            symbol = self.bars.get(bar_index, {}).get(time_index, EMPTY_BEAT_SYMBOL)
            all_bars.append(symbol)

        for i, bar in enumerate(chunked(all_bars, 4)):
            if i in self.bars:
                yield f"|{''.join(bar)}|"
            else:
                yield ""


class MemoDumpedSection(JubeatAnalyserDumpedSection):
    def render(self, circle_free: bool) -> str:
        blocs = []
        commands = list(self._dump_commands())
        if commands:
            blocs.append(commands)
        symbols = list(self._dump_symbol_definitions())
        if symbols:
            blocs.append(symbols)
        notes = list(self._dump_notes(circle_free))
        if notes:
            blocs.append(notes)
        return "\n".join(collapse(intersperse("", blocs)))

    def _dump_notes(self, circle_free: bool) -> Iterator[str]:
        notes_by_bar: Dict[int, List[AnyNote]] = defaultdict(list)
        bars: Dict[int, Dict[int, str]] = defaultdict(dict)
        chosen_symbols: Dict[BeatsTime, str] = {}
        symbols_iterator = iter(NOTE_SYMBOLS)
        for note in self.notes:
            time_in_section = note.time - self.current_beat
            bar_index = int(time_in_section)
            notes_by_bar[bar_index].append(note)
            if time_in_section % Fraction(1, 4) == 0:
                time_in_bar = time_in_section % Fraction(1)
                time_index = int(time_in_bar * 4)
                if time_index not in bars[bar_index]:
                    symbol = next(symbols_iterator)
                    chosen_symbols[time_in_section] = symbol
                    bars[bar_index][time_index] = symbol
            elif time_in_section not in self.symbols:
                raise ValueError(f"No symbol defined for time : {time_in_section}")

        # Create frame by bar
        section_symbols = ChainMap(chosen_symbols, self.symbols)
        frames_by_bar: Dict[int, List[Frame]] = defaultdict(list)
        for bar_index in range(ceil(self.length)):
            bar = bars.get(bar_index, {})
            frame = Frame()
            frame.bars[bar_index] = bar
            for note in notes_by_bar[bar_index]:
                time_in_section = note.time - self.current_beat
                symbol = section_symbols[time_in_section]
                if isinstance(note, TapNote):
                    if note.position in frame.positions:
                        frames_by_bar[bar_index].append(frame)
                        frame = Frame()
                    frame.positions[note.position] = symbol
                elif isinstance(note, LongNote):
                    needed_positions = set(note.positions_covered())
                    if needed_positions & frame.positions.keys():
                        frames_by_bar[bar_index].append(frame)
                        frame = Frame()
                    direction = note.tail_direction()
                    arrow = DIRECTION_TO_ARROW[direction]
                    line = DIRECTION_TO_LINE[direction]
                    for is_first, is_last, pos in mark_ends(note.positions_covered()):
                        if is_first:
                            frame.positions[pos] = symbol
                        elif is_last:
                            frame.positions[pos] = arrow
                        else:
                            frame.positions[pos] = line
                elif isinstance(note, LongNoteEnd):
                    if note.position in frame.positions:
                        frames_by_bar[bar_index].append(frame)
                        frame = Frame()
                    if circle_free and symbol in NOTE_TO_CIRCLE_FREE_SYMBOL:
                        symbol = NOTE_TO_CIRCLE_FREE_SYMBOL[symbol]
                    frame.positions[note.position] = symbol
            frames_by_bar[bar_index].append(frame)

        # Merge bar-specific frames is possible
        final_frames: List[Frame] = []
        for bar_index in range(ceil(self.length)):
            frames = frames_by_bar[bar_index]
            # Merge if :
            #  - No split in current bar (only one frame)
            #  - There is a previous frame
            #  - The previous frame is not a split frame (it holds a bar)
            #  - The previous and current bars are all in the same 4-bar group
            #  - The note positions in the previous frame do not clash with the current frame
            if (
                len(frames) == 1
                and final_frames
                and final_frames[-1].bars
                and max(final_frames[-1].bars.keys()) // 4
                == min(frames[0].bars.keys()) // 4
                and (
                    not (final_frames[-1].positions.keys() & frames[0].positions.keys())
                )
            ):
                final_frames[-1].bars.update(frames[0].bars)
                final_frames[-1].positions.update(frames[0].positions)
            else:
                final_frames.extend(frames)

        dumped_frames = map(lambda f: f.dump(self.length), final_frames)
        yield from collapse(intersperse("", dumped_frames))


def _raise_if_unfit_for_memo(chart: Chart, timing: Timing, circle_free: bool) -> None:
    if len(timing.events) < 1:
        raise ValueError("No BPM found in file") from None

    first_bpm = min(timing.events, key=lambda e: e.time)
    if first_bpm.time != 0:
        raise ValueError("First BPM event does not happen on beat zero")

    if any(
        not note.has_straight_tail()
        for note in chart.notes
        if isinstance(note, LongNote)
    ):
        raise ValueError(
            "Chart contains diagonal long notes, reprensenting these in"
            " memo format is not supported by jubeatools"
        )


def _section_factory(b: BeatsTime) -> MemoDumpedSection:
    return MemoDumpedSection(current_beat=b)


def _dump_memo_chart(
    difficulty: str,
    chart: Chart,
    metadata: Metadata,
    timing: Timing,
    circle_free: bool = False,
) -> StringIO:

    _raise_if_unfit_for_memo(chart, timing, circle_free)

    sections = create_sections_from_chart(
        MemoDumpedSection, chart, difficulty, timing, metadata, circle_free
    )

    # Jubeat Analyser format command
    sections[BeatsTime(0)].commands["memo"] = None

    # Define extra symbols
    existing_symbols: Dict[BeatsTime, str] = {}
    extra_symbols = iter(DEFAULT_EXTRA_SYMBOLS)
    for section_start, section in sections.items():
        # intentionally not a copy : at the end of this loop every section
        # holds a reference to a dict containing every defined symbol
        section.symbols = existing_symbols
        for note in section.notes:
            time_in_section = note.time - section_start
            if (
                time_in_section % Fraction(1, 4) != 0
                and time_in_section not in existing_symbols
            ):
                new_symbol = next(extra_symbols)
                section.symbol_definitions[time_in_section] = new_symbol
                existing_symbols[time_in_section] = new_symbol

    # Actual output to file
    file = StringIO()
    file.write(f"// Converted using jubeatools {__version__}\n")
    file.write(f"// {__repository_url__}\n\n")
    file.write(
        "\n\n".join(section.render(circle_free) for _, section in sections.items())
    )

    return file


dump_memo = make_full_dumper_from_jubeat_analyser_chart_dumper(_dump_memo_chart)



from typing import Any, Dict

from marshmallow import (
    RAISE,
    Schema,
    ValidationError,
    fields,
    validate,
    validates_schema,
)

# v0.x.x long note value :
#
#         8
#         4
#         0
#  11 7 3 . 1 5 9
#         2
#         6
#        10

X_Y_OFFSET_TO_P_VALUE = {
    (0, -1): 0,
    (0, -2): 4,
    (0, -3): 8,
    (0, 1): 2,
    (0, 2): 6,
    (0, 3): 10,
    (1, 0): 1,
    (2, 0): 5,
    (3, 0): 9,
    (-1, 0): 3,
    (-2, 0): 7,
    (-3, 0): 11,
}

P_VALUE_TO_X_Y_OFFSET = {v: k for k, v in X_Y_OFFSET_TO_P_VALUE.items()}


class StrictSchema(Schema):
    class Meta:
        ordered = True
        unknown = RAISE


class MemonNote(StrictSchema):
    n = fields.Integer(required=True, validate=validate.Range(min=0, max=15))
    t = fields.Integer(required=True, validate=validate.Range(min=0))
    # flake8 doesn't like "l" as a name here, so I silence the precise warning
    l = fields.Integer(required=True, validate=validate.Range(min=0))  # noqa: E741
    p = fields.Integer(required=True, validate=validate.Range(min=0, max=11))

    @validates_schema
    def validate_tail_tip_position(self, data: Dict[str, int], **kwargs: Any) -> None:
        if data["l"] > 0:
            x = data["n"] % 4
            y = data["n"] // 4
            dx, dy = P_VALUE_TO_X_Y_OFFSET[data["p"]]
            if not (0 <= x + dx < 4 and 0 <= y + dy < 4):
                raise ValidationError("Invalid tail position : {data}")


class MemonChart_0_1_0(StrictSchema):
    level = fields.Decimal(required=True)
    resolution = fields.Integer(required=True, validate=validate.Range(min=1))
    notes = fields.Nested(MemonNote, many=True)


class MemonChart_legacy(MemonChart_0_1_0):
    dif_name = fields.String(required=True)


class MemonMetadata_legacy(StrictSchema):
    title = fields.String(required=True, data_key="song title")
    artist = fields.String(required=True)
    audio = fields.String(required=True, data_key="music path")
    cover = fields.String(required=True, data_key="jacket path")
    BPM = fields.Decimal(
        required=True, validate=validate.Range(min=0, min_inclusive=False)
    )
    offset = fields.Decimal(required=True)


class MemonMetadata_0_1_0(MemonMetadata_legacy):
    cover = fields.String(required=True, data_key="album cover path")


class MemonPreview(StrictSchema):
    position = fields.Decimal(required=True, validate=validate.Range(min=0))
    length = fields.Decimal(
        required=True, validate=validate.Range(min=0, min_inclusive=False)
    )


class MemonMetadata_0_2_0(MemonMetadata_0_1_0):
    preview = fields.Nested(MemonPreview)


class MemonMetadata_0_3_0(MemonMetadata_0_2_0):
    audio = fields.String(required=False, data_key="music path")
    cover = fields.String(required=False, data_key="album cover path")
    preview_path = fields.String(data_key="preview path")


class Memon_legacy(StrictSchema):
    metadata = fields.Nested(MemonMetadata_legacy, required=True)
    data = fields.Nested(MemonChart_legacy, required=True, many=True)


class Memon_0_1_0(StrictSchema):
    version = fields.String(required=True, validate=validate.OneOf(["0.1.0"]))
    metadata = fields.Nested(MemonMetadata_0_1_0, required=True)
    data = fields.Dict(
        keys=fields.String(), values=fields.Nested(MemonChart_0_1_0), required=True
    )


class Memon_0_2_0(StrictSchema):
    version = fields.String(required=True, validate=validate.OneOf(["0.2.0"]))
    metadata = fields.Nested(MemonMetadata_0_2_0, required=True)
    data = fields.Dict(
        keys=fields.String(), values=fields.Nested(MemonChart_0_1_0), required=True
    )


class Memon_0_3_0(StrictSchema):
    version = fields.String(required=True, validate=validate.OneOf(["0.3.0"]))
    metadata = fields.Nested(MemonMetadata_0_3_0, required=True)
    data = fields.Dict(
        keys=fields.String(), values=fields.Nested(MemonChart_0_1_0), required=True
    )



from collections import defaultdict
from pathlib import Path
from typing import DefaultDict, List

from jubeatools import song
from jubeatools.formats.dump_tools import make_dumper_from_chart_file_dumper
from jubeatools.formats.filetypes import ChartFile
from jubeatools.utils import group_by

from .. import commons as konami
from ..commons import AnyNote
from ..dump_tools import make_events_from_chart
from . import construct


def _dump_jbsq(song: song.Song, **kwargs: dict) -> List[ChartFile]:
    res = []
    for dif, chart, timing, hakus in song.iter_charts():
        events = make_events_from_chart(chart.notes, timing, hakus)
        jbsq_chart = make_jbsq_chart(events, chart.notes)
        chart_bytes = construct.jbsq.build(jbsq_chart)
        res.append(ChartFile(chart_bytes, song, dif, chart))

    return res


dump_jbsq = make_dumper_from_chart_file_dumper(
    internal_dumper=_dump_jbsq, file_name_template=Path("seq_{difficulty:l}.jbsq")
)


def make_jbsq_chart(events: List[konami.Event], notes: List[AnyNote]) -> construct.JBSQ:
    jbsq_events = [convert_event_to_jbsq(e) for e in events]
    num_events = len(events)
    combo = compute_max_combo(notes)
    end_time = next(e for e in events if e.command == konami.Command.END).time
    first_note_time_in_beats = min((n.time for n in notes), default=0)
    starting_notes = [n for n in notes if n.time == first_note_time_in_beats]
    starting_buttons = sum(1 << n.position.index for n in starting_notes)
    first_note_time = min(
        (
            e.time
            for e in events
            if e.command in (konami.Command.PLAY, konami.Command.LONG)
        ),
        default=0,
    )
    densities = compute_density_graph(events, end_time)
    jbsq_chart = construct.JBSQ(
        num_events=num_events,
        combo=combo,
        end_time=end_time,
        starting_buttons=starting_buttons,
        start_time=first_note_time,
        density_graph=densities,
        events=jbsq_events,
    )
    jbsq_chart.magic = choose_magic_bytes(notes)
    return jbsq_chart


def convert_event_to_jbsq(event: konami.Event) -> construct.Event:
    return construct.Event(
        type_=construct.EventType[event.command.name],
        time_in_ticks=event.time,
        value=event.value,
    )


def compute_max_combo(notes: List[AnyNote]) -> int:
    notes_by_type = group_by(notes, type)
    tap_notes = len(notes_by_type[song.TapNote])
    long_notes = len(notes_by_type[song.LongNote])
    return tap_notes + 2 * long_notes


def compute_density_graph(events: List[konami.Event], end_time: int) -> List[int]:
    events_by_type = group_by(events, lambda e: e.command)
    buckets: DefaultDict[int, int] = defaultdict(int)
    for tap in events_by_type[konami.Command.PLAY]:
        bucket = int((tap.time / end_time) * 120)
        buckets[bucket] += 1

    for long in events_by_type[konami.Command.LONG]:
        press_bucket = int((long.time / end_time) * 120)
        buckets[press_bucket] += 1
        duration = konami.EveLong.from_value(long.value).duration
        release_time = long.time + duration
        release_bucket = int((release_time / end_time) * 120)
        buckets[release_bucket] += 1

    res = []
    for i in range(0, 120, 2):
        # The jbsq density graph in a array of nibbles, the twist is that for
        # some obscure reason each pair of nibbles is swapped in the byte ...
        # little-endianness is a hell of a drug, don't do drugs kids ...
        first_nibble = min(buckets[i], 15)
        second_nibble = min(buckets[i + 1], 15)
        density_byte = (second_nibble << 4) + first_nibble
        res.append(density_byte)

    return res


def choose_magic_bytes(notes: List[AnyNote]) -> bytes:
    # The magic bytes are actually checked by the app to choose whether or not
    # it should display the "hold notes" symbol next to the song jacket
    if any(isinstance(note, song.LongNote) for note in notes):
        return b"IJSQ"
    else:
        return b"JBSQ"



from decimal import Decimal

from hypothesis import strategies as st

from jubeatools import song
from jubeatools.testutils import strategies as jbst

simple_beat_strat = jbst.beat_time(
    denominator_strat=st.sampled_from([4, 3]), max_section=10
)


@st.composite
def eve_compatible_song(draw: st.DrawFn) -> song.Song:
    """eve only keeps notes, hakus, timing info and difficulty,
    the precision you can get out of it is also severly limited"""
    diff = draw(st.sampled_from(list(d.value for d in song.Difficulty)))
    chart = draw(
        jbst.chart(
            timing_strat=jbst.timing_info(
                with_bpm_changes=True,
                bpm_strat=st.decimals(min_value=50, max_value=300, places=2),
                beat_zero_offset_strat=st.decimals(min_value=0, max_value=20, places=2),
                time_strat=jbst.beat_time(
                    min_section=1,
                    max_section=10,
                    denominator_strat=st.sampled_from([4, 3]),
                ),
            ),
            notes_strat=jbst.notes(
                note_strat=st.one_of(
                    jbst.tap_note(time_strat=simple_beat_strat),
                    jbst.long_note(
                        time_strat=simple_beat_strat,
                        duration_strat=jbst.beat_time(
                            min_numerator=1,
                            max_section=3,
                            denominator_strat=st.sampled_from([4, 3]),
                        ),
                    ),
                ),
                beat_time_strat=simple_beat_strat,
            ),
            level_strat=st.just(Decimal(0)),
            hakus_strat=st.one_of(st.none(), st.sets(simple_beat_strat)),
        )
    )
    return song.Song(
        metadata=song.Metadata(),
        charts={diff: chart},
    )



import pytest

from jubeatools.testutils.resources import load_test_file

from . import data


def test_RorataJins_example() -> None:
    """This file has a #memo tag but actually uses mono-column formatting,
    Here I just check that a friendlier error message is sent because there
    is not much else I can to do here, the file is plain old wrong"""
    with pytest.raises(SyntaxError, match="separator line"):
        load_test_file(data, "RorataJin's example.txt")


def test_Booths_of_Fighters_memo() -> None:
    """This file has 2 quirks my code did not anticipate :
    - while it's in #memo2 format, it actually uses b= and t= commands
    - the position and timing parts are separated by some less common
      whitespace character"""
    load_test_file(data, "Booths_of_Fighters_memo.txt")


def test_MTC_Nageki_no_Ki_EXT() -> None:
    """This file is proper euc-kr text that's specially crafted to also be
    compatible with jubeat analyser, handcrafted mojibake and all"""
    load_test_file(data, "MTC_Nageki_no_Ki_EXT.txt")


def test_MTC_Mimi_EXT() -> None:
    """Also an euc-kr file but also has long notes"""
    load_test_file(data, "MTC_Mimi_EXT.txt")


def test_beat_1() -> None:
    """This file uses #bpp=2 with half-width hyphens in the timing part, this
    is bound to fail on the first line that has an odd length when encoded
    to shift-jis, for now I just want to display a friendlier error message"""
    with pytest.raises(SyntaxError, match="Series of notes of odd byte lengths"):
        load_test_file(data, "beat(1).txt")



from decimal import Decimal
from pathlib import Path

from jubeatools.song import *

data = (
    Song(
        metadata=Metadata(
            title="",
            artist="",
            audio=Path(""),
            cover=Path(""),
            preview=None,
            preview_file=None,
        ),
        charts={
            "BSC": Chart(
                level=Decimal("0.0"),
                timing=Timing(
                    events=[
                        BPMEvent(time=Fraction(0, 1), BPM=Decimal("1.000")),
                        BPMEvent(time=Fraction(8, 1), BPM=Decimal("1.000")),
                    ],
                    beat_zero_offset=Decimal("0.000"),
                ),
                notes=[],
            )
        },
        common_timing=None,
    ),
    False,
)



import math
from fractions import Fraction
from functools import singledispatch
from typing import List, Optional, Set

from more_itertools import numeric_range

from jubeatools import song
from jubeatools.formats.timemap import TimeMap

from .commons import AnyNote, Command, Event, bpm_to_value, ticks_at_beat


def make_events_from_chart(
    notes: List[AnyNote], timing: song.Timing, hakus: Optional[Set[song.BeatsTime]]
) -> List[Event]:
    time_map = TimeMap.from_timing(timing)
    note_events = make_note_events(notes, time_map)
    timing_events = make_timing_events(notes, timing, hakus, time_map)
    return sorted(note_events + timing_events)


def make_note_events(notes: List[AnyNote], time_map: TimeMap) -> List[Event]:
    return [make_note_event(note, time_map) for note in notes]


@singledispatch
def make_note_event(note: AnyNote, time_map: TimeMap) -> Event:
    raise NotImplementedError(f"Unknown note type : {type(note)}")


@make_note_event.register
def make_tap_note_event(note: song.TapNote, time_map: TimeMap) -> Event:
    return Event.from_tap_note(note, time_map)


@make_note_event.register
def make_long_note_event(note: song.LongNote, time_map: TimeMap) -> Event:
    return Event.from_long_note(note, time_map)


def make_timing_events(
    notes: List[AnyNote],
    timing: song.Timing,
    hakus: Optional[Set[song.BeatsTime]],
    time_map: TimeMap,
) -> List[Event]:
    bpm_events = [make_bpm_event(e, time_map) for e in timing.events]
    end_beat = choose_end_beat(notes, hakus)
    end_event = make_end_event(end_beat, time_map)
    measure_events = make_measure_events(end_beat, time_map)
    if hakus is not None:
        haku_events = dump_hakus(hakus, time_map)
    else:
        haku_events = make_regular_hakus(end_beat, time_map)

    return bpm_events + measure_events + haku_events + [end_event]


def make_bpm_event(bpm_change: song.BPMEvent, time_map: TimeMap) -> Event:
    ticks = ticks_at_beat(bpm_change.time, time_map)
    bpm_value = math.floor(bpm_to_value(Fraction(bpm_change.BPM)))
    return Event(time=ticks, command=Command.TEMPO, value=bpm_value)


def choose_end_beat(
    notes: List[AnyNote], hakus: Optional[Set[song.BeatsTime]]
) -> song.BeatsTime:
    """Leave 2 empty measures (4 beats) after the last event"""
    last_note_beat = compute_last_note_beat(notes)
    last_haku_beat = max(hakus or [], default=song.BeatsTime(0))
    last_beat = max(last_haku_beat, last_note_beat)
    measure = last_beat - (last_beat % 4)
    return measure + song.BeatsTime(2 * 4)


def compute_last_note_beat(notes: List[AnyNote]) -> song.BeatsTime:
    """Returns the last beat at which a note event happens, either a tap note,
    the start of a long note or the end of a long note.

    If we don't take long notes ends into account we might end up with a long
    note end happening after the END tag which will cause jubeat to freeze when
    trying to render the note density graph"""
    note_times = set(n.time for n in notes)
    long_note_ends = set(
        n.time + n.duration for n in notes if isinstance(n, song.LongNote)
    )
    all_note_times = note_times | long_note_ends
    return max(all_note_times, default=song.BeatsTime(0))


def make_end_event(end_beat: song.BeatsTime, time_map: TimeMap) -> Event:
    ticks = ticks_at_beat(end_beat, time_map)
    return Event(time=ticks, command=Command.END, value=0)


def make_measure_events(end_beat: song.BeatsTime, time_map: TimeMap) -> List[Event]:
    start = song.BeatsTime(0)
    stop = end_beat + song.BeatsTime(1)
    step = song.BeatsTime(4)
    beats = numeric_range(start, stop, step)
    return [make_measure_event(beat, time_map) for beat in beats]


def make_measure_event(beat: song.BeatsTime, time_map: TimeMap) -> Event:
    ticks = ticks_at_beat(beat, time_map)
    return Event(time=ticks, command=Command.MEASURE, value=0)


def dump_hakus(hakus: Set[song.BeatsTime], time_map: TimeMap) -> List[Event]:
    return [make_haku_event(beat, time_map) for beat in sorted(hakus)]


def make_regular_hakus(end_beat: song.BeatsTime, time_map: TimeMap) -> List[Event]:
    start = song.BeatsTime(0)
    stop = end_beat + song.BeatsTime(1, 2)
    step = song.BeatsTime(1)
    beats = numeric_range(start, stop, step)
    return [make_haku_event(beat, time_map) for beat in beats]


def make_haku_event(beat: song.BeatsTime, time_map: TimeMap) -> Event:
    ticks = ticks_at_beat(beat, time_map)
    return Event(time=ticks, command=Command.HAKU, value=0)



"""
Provides the :py:obj:`Song` class, the central model for chartsets

- Every input format is converted to a :py:obj:`Song` instance
- Every output format is created from a :py:obj:`Song` instance

Most timing-related info is stored as beat fractions, otherwise a decimal
number of seconds is used"""

from __future__ import annotations

from collections import Counter
from dataclasses import Field, astuple, dataclass, field, fields
from decimal import Decimal
from enum import Enum, auto
from fractions import Fraction
from functools import wraps
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
)

from jubeatools.utils import none_or

#: A time measured in beats
BeatsTime = Fraction
#: A time measured in seconds
SecondsTime = Decimal


def beats_time_from_ticks(ticks: int, resolution: int) -> BeatsTime:
    if resolution < 1:
        raise ValueError(f"resolution cannot be negative : {resolution}")
    return BeatsTime(ticks, resolution)


def convert_other(
    f: Callable[[Position, Position], Position]
) -> Callable[[Position, Any], Position]:
    @wraps(f)
    def wrapped(self: Position, other: Any) -> Position:
        if isinstance(other, Position):
            other_pos = other
        else:
            try:
                other_pos = Position(*other)
            except Exception:
                raise ValueError(f"Could not convert {type(other)} to a Position")

        return f(self, other_pos)

    return wrapped


@dataclass(frozen=True, order=True)
class Position:
    """2D integer vector"""

    x: int
    y: int

    def __iter__(self) -> Iterator[int]:
        yield from astuple(self)

    @convert_other
    def __add__(self, other: Position) -> Position:
        return Position(self.x + other.x, self.y + other.y)

    @convert_other
    def __sub__(self, other: Position) -> Position:
        return Position(self.x - other.x, self.y - other.y)

    def __mul__(self, other: int) -> Position:
        return Position(self.x * other, self.y * other)

    __rmul__ = __mul__


@dataclass(frozen=True, order=True)
class NotePosition(Position):
    """
    A specific square on the controller. (0, 0) is the top-left button, x
    goes right, y goes down.

    ::

            x →
            0 1 2 3
        y 0 □ □ □ □
        ↓ 1 □ □ □ □
          2 □ □ □ □
          3 □ □ □ □

    The main difference with Position is that x and y MUST be between 0 and 3
    """

    def __post_init__(self) -> None:
        if not 0 <= self.x < 4:
            raise ValueError("x out of [0, 3] range")
        if not 0 <= self.y < 4:
            raise ValueError("y out of [0, 3] range")

    @property
    def index(self) -> int:
        return self.x + 4 * self.y

    @classmethod
    def from_index(cls, index: int) -> NotePosition:
        if not (0 <= index < 16):
            raise ValueError(f"Note position index out of range : {index}")

        return cls(x=index % 4, y=index // 4)

    @classmethod
    def from_raw_position(cls, pos: Position) -> NotePosition:
        return cls(x=pos.x, y=pos.y)


@dataclass(frozen=True, unsafe_hash=True)
class TapNote:
    time: BeatsTime
    position: NotePosition


@dataclass(frozen=True, unsafe_hash=True)
class LongNote:
    time: BeatsTime
    position: NotePosition
    duration: BeatsTime
    # tail tip starting position as absolute position on the playfield
    tail_tip: NotePosition

    def has_straight_tail(self) -> bool:
        return (self.position.x == self.tail_tip.x) or (
            self.position.y == self.tail_tip.y
        )

    def tail_direction(self) -> Direction:
        """Direction in which the tail moves"""
        if not self.has_straight_tail():
            raise ValueError("Can't get cardinal direction of diagonal long note")

        if self.tail_tip.x == self.position.x:
            if self.tail_tip.y > self.position.y:
                return Direction.UP
            else:
                return Direction.DOWN
        else:
            if self.tail_tip.x > self.position.x:
                return Direction.LEFT
            else:
                return Direction.RIGHT

    def positions_covered(self) -> Iterator[NotePosition]:
        direction = self.tail_direction()
        step = TAIL_DIRECTION_TO_OUTWARDS_VECTOR[direction]
        position = self.position
        yield position
        while position != self.tail_tip:
            position = NotePosition.from_raw_position(position + step)
            yield position


class Direction(Enum):
    UP = auto()
    DOWN = auto()
    LEFT = auto()
    RIGHT = auto()


TAIL_DIRECTION_TO_OUTWARDS_VECTOR = {
    Direction.UP: Position(0, 1),
    Direction.DOWN: Position(0, -1),
    Direction.LEFT: Position(1, 0),
    Direction.RIGHT: Position(-1, 0),
}


@dataclass(frozen=True)
class BPMEvent:
    time: BeatsTime
    BPM: Decimal


@dataclass(unsafe_hash=True)
class Timing:
    events: Sequence[BPMEvent]
    beat_zero_offset: SecondsTime

    def __post_init__(self) -> None:
        self.events = tuple(self.events)


@dataclass
class Chart:
    level: Optional[Decimal]
    timing: Optional[Timing] = None
    hakus: Optional[Set[BeatsTime]] = None
    notes: List[Union[TapNote, LongNote]] = field(default_factory=list)


@dataclass(frozen=True)
class Preview:
    """Frozen so it can be hashed to be deduped using a set when merging
    Metadata instances"""

    start: SecondsTime
    length: SecondsTime


@dataclass
class Metadata:
    title: Optional[str] = None
    artist: Optional[str] = None
    audio: Optional[Path] = None
    cover: Optional[Path] = None
    preview: Optional[Preview] = None
    preview_file: Optional[Path] = None

    @classmethod
    def permissive_merge(cls, *metadatas: "Metadata") -> "Metadata":
        """Make the "sum" of all the given metadata instances, if possible. If
        several instances have different defined values for the same field,
        merging will fail. Fields with Noneor empty values (empty string or
        empty path) are conscidered undefined and their values can be replaced
        by an actual value if supplied by at least one object from the given
        iterable."""
        return cls(
            **{f.name: _get_common_value(f, metadatas) for f in fields(cls)},
        )


def _get_common_value(field_: Field, metadatas: Iterable[Metadata]) -> Any:
    raw_values = []
    empty_values = []
    for m in metadatas:
        value = getattr(m, field_.name)
        if value is None:
            continue
        elif not value:
            empty_values.append(value)
        else:
            raw_values.append(value)

    real_values = set(raw_values)
    if len(real_values) > 1:
        raise ValueError(
            f"Can't merge metadata, the {field_.name} field has "
            f"conflicting possible values : {real_values}"
        )

    try:
        return real_values.pop()
    except KeyError:
        pass

    try:
        return empty_values.pop()
    except IndexError:
        pass

    return None


class Difficulty(str, Enum):
    BASIC = "BSC"
    ADVANCED = "ADV"
    EXTREME = "EXT"

    def __str__(self) -> str:
        return self.value


@dataclass
class Song:
    """The abstract representation format for all jubeat chart sets.
    A Song is a set of charts with associated metadata"""

    #: Miscellaneous information about the song
    metadata: Metadata
    #: A regular dict that maps a difficulty name to a :py:obj:`Chart`.
    #: The names for the usual jubeat difficulties should be ``"BSC"``, ``"ADV"``,
    #: and ``"EXT"``
    charts: Dict[str, Chart] = field(default_factory=dict)
    #: The optional shared timing object that applies to all charts by default
    common_timing: Optional[Timing] = None
    #: The optional shared set of HAKUs that apply to all charts by default
    common_hakus: Optional[Set[BeatsTime]] = None

    @classmethod
    def from_monochart_instances(cls, *songs: "Song") -> "Song":
        metadata = Metadata.permissive_merge(*(song.metadata for song in songs))
        charts: Dict[str, Chart] = {}
        for song in songs:
            song.remove_common_timing()
            song.remove_common_hakus()
            charts.update(song.charts)

        merged = cls(
            metadata=metadata,
            charts=charts,
        )
        merged.minimize_timings()
        merged.minimize_hakus()
        return merged

    def minimize_timings(self) -> None:
        """Turn timings into minimal form : Use the most common timing object
        as the common timing, if it is used by more than two charts, otherwise
        each chart gets its own timing object and no common timing is defined"""
        self.remove_common_timing()
        counts = Counter(c.timing for c in self.charts.values())
        ((most_used, count),) = counts.most_common(1)
        if count >= 2:
            self.common_timing = most_used
            for chart in self.charts.values():
                if chart.timing == most_used:
                    chart.timing = None

    def remove_common_timing(self) -> None:
        """Modify the song object so that no chart relies on the common timing
        object, charts that previously did rely on it now have a chart-specific
        timing object equal to the old common timing"""
        for _, chart, applicable_timing in self.iter_charts_with_applicable_timing():
            chart.timing = applicable_timing

        self.common_timing = None

    def minimize_hakus(self) -> None:
        """Same deal as "minimize_timings" but with hakus"""
        self.remove_common_hakus()
        counts = Counter(
            none_or(frozenset.__call__, c.hakus) for c in self.charts.values()
        )
        most_used, count = counts.most_common(1).pop()
        if count >= 2:
            self.common_hakus = most_used
            for chart in self.charts.values():
                if chart.hakus == most_used:
                    chart.hakus = None

    def remove_common_hakus(self) -> None:
        for chart in self.charts.values():
            if chart.hakus is None:
                chart.hakus = self.common_hakus

        self.common_hakus = None

    def iter_charts_with_applicable_timing(self) -> Iterator[Tuple[str, Chart, Timing]]:
        for dif, chart in self.charts.items():
            timing = chart.timing or self.common_timing
            if timing is None:
                raise ValueError(
                    f"Neither song nor {dif} chart have any timing information"
                )
            yield dif, chart, timing

    def iter_charts(
        self,
    ) -> Iterator[Tuple[str, Chart, Timing, Optional[Set[BeatsTime]]]]:
        for dif, chart in self.charts.items():
            timing = chart.timing or self.common_timing
            if timing is None:
                raise ValueError(
                    f"Neither song nor {dif} chart have any timing information"
                )
            hakus = chart.hakus if chart.hakus is not None else self.common_hakus
            yield dif, chart, timing, hakus



from dataclasses import astuple, dataclass, field
from decimal import Decimal
from functools import partial
from typing import Any, Dict, List, Optional, Tuple, Union

from marshmallow import EXCLUDE, Schema, ValidationError, post_dump, validate
from marshmallow_dataclass import NewType, class_schema

PositiveDecimal = NewType("PositiveDecimal", Decimal, validate=validate.Range(min=0))
StrictlyPositiveDecimal = NewType(
    "StrictlyPositiveDecimal",
    Decimal,
    validate=validate.Range(min=0, min_inclusive=False),
)


@dataclass
class PreviewSample:
    start: PositiveDecimal
    duration: StrictlyPositiveDecimal


Preview = Union[str, PreviewSample]


@dataclass
class Metadata:
    title: Optional[str]
    artist: Optional[str]
    audio: Optional[str]
    jacket: Optional[str]
    preview: Optional[Preview]


PositiveInt = NewType("PositiveInt", int, validate=validate.Range(min=0))
StrictlyPositiveInt = NewType(
    "StrictlyPositiveInteger", int, validate=validate.Range(min=0, min_inclusive=False)
)
MixedNumber = Tuple[int, int, int]
SymbolicTime = Union[int, MixedNumber]


def validate_symbolic_time(t: SymbolicTime) -> None:
    if isinstance(t, int):
        if t < 0:
            raise ValidationError("Negative ticks are not allowed")
    elif isinstance(t, tuple):
        validate_mixed_number(t)


def validate_mixed_number(m: MixedNumber) -> None:
    if m[0] < 0:
        raise ValidationError("First number in fraction tuple can't be negative")
    elif m[1] < 0:
        raise ValidationError("Second number in fraction tuple can't be negative")
    elif m[2] < 1:
        raise ValidationError("Third number in fraction tuple can't be less than 1")


def validate_strictly_positive_mixed_number(m: MixedNumber) -> None:
    if (m[0], m[1]) == (0, 0):
        raise ValidationError("The tuple must represent a strictly positive number")


@dataclass
class BPMEvent:
    beat: SymbolicTime = field(metadata={"validate": validate_symbolic_time})
    bpm: StrictlyPositiveDecimal


@dataclass
class Timing:
    offset: Optional[Decimal]
    resolution: Optional[StrictlyPositiveInt]
    bpms: Optional[List[BPMEvent]]
    hakus: Optional[List[SymbolicTime]] = field(
        metadata={"validate": partial(map, validate_symbolic_time)}
    )

    def remove_default_values(self) -> "Timing":
        return self.remove_common_values(DEFAULT_TIMING)

    def remove_common_values(self, other: "Timing") -> "Timing":
        return Timing(
            offset=None if self.offset == other.offset else self.offset,
            resolution=None if self.resolution == other.resolution else self.resolution,
            bpms=None if self.bpms == other.bpms else self.bpms,
            hakus=None if self.hakus == other.hakus else self.hakus,
        )

    def __bool__(self) -> bool:
        return any(f is not None for f in astuple(self))

    @classmethod
    def fill_in_defaults(cls, *timings: Optional["Timing"]) -> "Timing":
        ordered = [*timings, DEFAULT_TIMING]
        return cls.merge(*(t for t in ordered if t is not None))

    @classmethod
    def merge(cls, *timings: "Timing") -> "Timing":
        offset = next((t.offset for t in timings if t.offset is not None), None)
        resolution = next(
            (t.resolution for t in timings if t.resolution is not None), None
        )
        bpms = next((t.bpms for t in timings if t.bpms is not None), None)
        hakus = next((t.hakus for t in timings if t.hakus is not None), None)
        return cls(
            offset=offset,
            resolution=resolution,
            bpms=bpms,
            hakus=hakus,
        )


DEFAULT_TIMING = Timing(
    offset=Decimal(0),
    resolution=240,
    bpms=[BPMEvent(0, Decimal(120))],
    hakus=None,
)

Button = NewType("Button", int, validate=validate.Range(min=0, max=15))


@dataclass
class TapNote:
    n: Button
    t: SymbolicTime = field(metadata={"validate": validate_symbolic_time})


def validate_symbolic_duration(t: SymbolicTime) -> None:
    if isinstance(t, int):
        if t < 1:
            raise ValidationError("Duration has to be positive and non-zero")
    elif isinstance(t, tuple):
        validate_mixed_number(t)
        validate_strictly_positive_mixed_number(t)


TailIn6Notation = NewType("TailIn6Notation", int, validate=validate.Range(min=0, max=5))


@dataclass
class LongNote(TapNote):
    l: SymbolicTime = field(metadata={"validate": validate_symbolic_duration})
    p: TailIn6Notation


# LongNote first otherwise long notes get interpreted as tap notes
Note = Union[LongNote, TapNote]


@dataclass
class Chart:
    level: Optional[Decimal]
    resolution: Optional[StrictlyPositiveInt]
    timing: Optional[Timing]
    notes: List[Note]


Version = NewType("Version", str, validate=validate.Equal("1.0.0"))


@dataclass
class File:
    version: Version
    metadata: Optional[Metadata]
    timing: Optional[Timing]
    data: Dict[str, Chart]


class BaseSchema(Schema):
    class Meta:
        ordered = True
        unknown = EXCLUDE

    @post_dump
    def _remove_none_values(self, data: dict, **kwargs: Any) -> dict:
        return remove_none_values(data)


def remove_none_values(data: dict) -> dict:
    return {key: value for key, value in data.items() if value is not None}


FILE_SCHEMA = class_schema(File, base_schema=BaseSchema)()



from pathlib import Path
from typing import Any, Dict, Union

from jubeatools import song as jbt
from jubeatools.utils import none_or

from ..tools import make_memon_folder_loader
from . import schema as memon


def _load_memon_note_v0(
    note: dict, resolution: int
) -> Union[jbt.TapNote, jbt.LongNote]:
    position = jbt.NotePosition.from_index(note["n"])
    time = jbt.beats_time_from_ticks(ticks=note["t"], resolution=resolution)
    if note["l"] > 0:
        duration = jbt.beats_time_from_ticks(ticks=note["l"], resolution=resolution)
        p_value = note["p"]
        𝛿x, 𝛿y = memon.P_VALUE_TO_X_Y_OFFSET[p_value]
        tail_tip = jbt.NotePosition.from_raw_position(position + jbt.Position(𝛿x, 𝛿y))
        return jbt.LongNote(time, position, duration, tail_tip)
    else:
        return jbt.TapNote(time, position)


def _load_memon_legacy(raw_memon: Any) -> jbt.Song:
    schema = memon.Memon_legacy()
    file = schema.load(raw_memon)
    metadata = jbt.Metadata(
        title=file["metadata"]["title"],
        artist=file["metadata"]["artist"],
        audio=Path(file["metadata"]["audio"]),
        cover=Path(file["metadata"]["cover"]),
    )
    common_timing = jbt.Timing(
        events=[jbt.BPMEvent(time=jbt.BeatsTime(0), BPM=file["metadata"]["BPM"])],
        beat_zero_offset=jbt.SecondsTime(-file["metadata"]["offset"]),
    )
    charts: Dict[str, jbt.Chart] = {}
    for memon_chart in file["data"]:
        difficulty = memon_chart["dif_name"]
        chart = jbt.Chart(
            level=memon_chart["level"],
            notes=[
                _load_memon_note_v0(note, memon_chart["resolution"])
                for note in memon_chart["notes"]
            ],
        )
        charts[difficulty] = chart

    return jbt.Song(metadata=metadata, charts=charts, common_timing=common_timing)


load_memon_legacy = make_memon_folder_loader(_load_memon_legacy)


def _load_memon_0_1_0(raw_memon: Any) -> jbt.Song:
    schema = memon.Memon_0_1_0()
    file = schema.load(raw_memon)
    metadata = jbt.Metadata(
        title=file["metadata"]["title"],
        artist=file["metadata"]["artist"],
        audio=Path(file["metadata"]["audio"]),
        cover=Path(file["metadata"]["cover"]),
    )
    common_timing = jbt.Timing(
        events=[jbt.BPMEvent(time=jbt.BeatsTime(0), BPM=file["metadata"]["BPM"])],
        beat_zero_offset=jbt.SecondsTime(-file["metadata"]["offset"]),
    )
    charts: Dict[str, jbt.Chart] = {}
    for difficulty, memon_chart in file["data"].items():
        chart = jbt.Chart(
            level=memon_chart["level"],
            notes=[
                _load_memon_note_v0(note, memon_chart["resolution"])
                for note in memon_chart["notes"]
            ],
        )
        charts[difficulty] = chart

    return jbt.Song(metadata=metadata, charts=charts, common_timing=common_timing)


load_memon_0_1_0 = make_memon_folder_loader(_load_memon_0_1_0)


def _load_memon_0_2_0(raw_memon: Any) -> jbt.Song:
    schema = memon.Memon_0_2_0()
    file = schema.load(raw_memon)
    preview = None
    if "preview" in file["metadata"]:
        start = file["metadata"]["preview"]["position"]
        length = file["metadata"]["preview"]["length"]
        preview = jbt.Preview(start, length)

    metadata = jbt.Metadata(
        title=file["metadata"]["title"],
        artist=file["metadata"]["artist"],
        audio=Path(file["metadata"]["audio"]),
        cover=Path(file["metadata"]["cover"]),
        preview=preview,
    )
    common_timing = jbt.Timing(
        events=[jbt.BPMEvent(time=jbt.BeatsTime(0), BPM=file["metadata"]["BPM"])],
        beat_zero_offset=jbt.SecondsTime(-file["metadata"]["offset"]),
    )
    charts: Dict[str, jbt.Chart] = {}
    for difficulty, memon_chart in file["data"].items():
        chart = jbt.Chart(
            level=memon_chart["level"],
            notes=[
                _load_memon_note_v0(note, memon_chart["resolution"])
                for note in memon_chart["notes"]
            ],
        )
        charts[difficulty] = chart

    return jbt.Song(metadata=metadata, charts=charts, common_timing=common_timing)


load_memon_0_2_0 = make_memon_folder_loader(_load_memon_0_2_0)


def _load_memon_0_3_0(raw_memon: Any) -> jbt.Song:
    schema = memon.Memon_0_3_0()
    file = schema.load(raw_memon)
    preview = None
    if "preview" in file["metadata"]:
        start = file["metadata"]["preview"]["position"]
        length = file["metadata"]["preview"]["length"]
        preview = jbt.Preview(start, length)

    metadata = jbt.Metadata(
        title=file["metadata"]["title"],
        artist=file["metadata"]["artist"],
        audio=none_or(Path, file["metadata"].get("audio")),
        cover=none_or(Path, file["metadata"].get("cover")),
        preview=preview,
        preview_file=none_or(Path, file["metadata"].get("preview_path")),
    )
    common_timing = jbt.Timing(
        events=[jbt.BPMEvent(time=jbt.BeatsTime(0), BPM=file["metadata"]["BPM"])],
        beat_zero_offset=jbt.SecondsTime(-file["metadata"]["offset"]),
    )
    charts: Dict[str, jbt.Chart] = {}
    for difficulty, memon_chart in file["data"].items():
        chart = jbt.Chart(
            level=memon_chart["level"],
            notes=[
                _load_memon_note_v0(note, memon_chart["resolution"])
                for note in memon_chart["notes"]
            ],
        )
        charts[difficulty] = chart

    return jbt.Song(metadata=metadata, charts=charts, common_timing=common_timing)


load_memon_0_3_0 = make_memon_folder_loader(_load_memon_0_3_0)



"""Command Line Interface"""

from pathlib import Path
from typing import Any, Dict, Optional

import click

from jubeatools.formats import DUMPERS, LOADERS, Format
from jubeatools.formats.guess import guess_format

from .helpers import dumper_option, loader_option


@click.command()
@click.argument("src", type=click.Path(exists=True, dir_okay=True))
@click.argument("dst", type=click.Path())
@click.option(
    "--input-format",
    "input_format",
    type=click.Choice(list(f._value_ for f in LOADERS.keys())),
    help=(
        "Force jubeatools to read the input file/folder as the given format."
        "If this option is not used jubeatools will try to guess the format"
    ),
)
@click.option(
    "-f",
    "--format",
    "output_format",
    required=True,
    prompt="Choose an output format",
    type=click.Choice(list(f._value_ for f in DUMPERS.keys())),
    help="Output file format",
)
@dumper_option(
    "--circlefree",
    "circle_free",
    is_flag=True,
    help="Use #circlefree=1 for jubeat analyser formats",
)
@loader_option(
    "--beat-snap",
    "beat_snap",
    type=click.IntRange(min=1),
    help=(
        "For compatible input formats, snap all notes and bpm changes to "
        "the nearest 1/beat_snap beat"
    ),
)
@loader_option(
    "--merge",
    is_flag=True,
    help="For memon, if called on a folder, merge all the .memon files found",
)
def convert(
    src: str,
    dst: str,
    input_format: Optional[Format],
    output_format: Format,
    loader_options: Optional[Dict[str, Any]] = None,
    dumper_options: Optional[Dict[str, Any]] = None,
) -> None:
    """Convert SRC to DST using the format specified by -f"""
    if input_format is None:
        input_format = guess_format(Path(src))
        click.echo(f"Detected input file format : {input_format}")

    try:
        loader = LOADERS[input_format]
    except KeyError:
        raise ValueError(f"Unsupported input format : {input_format}")

    try:
        dumper = DUMPERS[output_format]
    except KeyError:
        raise ValueError(f"Unsupported output format : {input_format}")

    loader_options = loader_options or {}
    song = loader(Path(src), **loader_options)
    dumper_options = dumper_options or {}
    files = dumper(song, Path(dst), **dumper_options)
    for path, contents in files.items():
        with path.open("wb") as f:
            f.write(contents)


if __name__ == "__main__":
    convert()



import string
from functools import singledispatch
from itertools import count
from pathlib import Path
from typing import AbstractSet, Any, Dict, Iterator, Optional, TypedDict

from jubeatools.formats.filetypes import ChartFile, JubeatFile, SongFile
from jubeatools.formats.typing import ChartFileDumper, Dumper
from jubeatools.song import Difficulty, Song
from jubeatools.utils import none_or

DIFFICULTY_NUMBER: Dict[str, int] = {
    Difficulty.BASIC: 1,
    Difficulty.ADVANCED: 2,
    Difficulty.EXTREME: 3,
}

DIFFICULTY_INDEX: Dict[str, int] = {
    Difficulty.BASIC: 0,
    Difficulty.ADVANCED: 1,
    Difficulty.EXTREME: 2,
}


def make_dumper_from_chart_file_dumper(
    internal_dumper: ChartFileDumper,
    file_name_template: Path,
) -> Dumper:
    """Adapt a ChartFileDumper to the Dumper protocol, The resulting function
    uses the file name template if it recieves an existing directory as an
    output path"""

    def dumper(song: Song, path: Path, **kwargs: Any) -> Dict[Path, bytes]:
        res: Dict[Path, bytes] = {}
        name_format = FileNameFormat(file_name_template, suggestion=path)
        files = internal_dumper(song, **kwargs)
        for chartfile in files:
            filepath = name_format.available_filename_for(
                chartfile, already_chosen=res.keys()
            )
            res[filepath] = chartfile.contents

        return res

    return dumper


class FormatParameters(TypedDict, total=False):
    title: str
    # uppercase BSC ADV EXT
    difficulty: str
    # 0-based
    difficulty_index: str
    # 1-based
    difficulty_number: str
    dedup: str


class FileNameFormat:
    def __init__(self, file_name_template: Path, suggestion: Path):
        if suggestion.is_dir():
            file_path = file_name_template
            self.parent = suggestion
        else:
            file_path = suggestion
            self.parent = suggestion.parent

        self.name_format = f"{file_path.stem}{{dedup}}{file_path.suffix}"

    def available_filename_for(
        self, file: JubeatFile, already_chosen: Optional[AbstractSet[Path]] = None
    ) -> Path:
        fixed_params = extract_format_params(file)
        return next(self.iter_possible_paths(fixed_params, already_chosen))

    def iter_possible_paths(
        self,
        fixed_params: FormatParameters,
        already_chosen: Optional[AbstractSet[Path]] = None,
    ) -> Iterator[Path]:
        all_paths = self.iter_deduped_paths(fixed_params)
        not_on_filesystem = (p for p in all_paths if not p.exists())
        if already_chosen is not None:
            yield from (p for p in not_on_filesystem if p not in already_chosen)
        else:
            yield from not_on_filesystem

    def iter_deduped_paths(self, params: FormatParameters) -> Iterator[Path]:
        for dedup_index in count(start=0):
            # TODO Remove the type ignore once this issue is fixed
            # https://github.com/python/mypy/issues/6019
            params.update(  # type: ignore[call-arg]
                dedup="" if dedup_index == 0 else f"-{dedup_index}"
            )
            formatter = BetterStringFormatter()
            filename = formatter.format(self.name_format, **params).strip()
            yield self.parent / filename


@singledispatch
def extract_format_params(file: JubeatFile) -> FormatParameters:
    raise ValueError(f"Unexpected argument : {file}")


@extract_format_params.register
def extract_song_format_params(songfile: SongFile) -> FormatParameters:
    return FormatParameters(title=none_or(slugify, songfile.song.metadata.title) or "")


@extract_format_params.register
def extract_chart_format_params(chartfile: ChartFile) -> FormatParameters:
    return FormatParameters(
        title=none_or(slugify, chartfile.song.metadata.title) or "",
        difficulty=slugify(chartfile.difficulty),
        difficulty_index=str(DIFFICULTY_INDEX.get(chartfile.difficulty, 2)),
        difficulty_number=str(DIFFICULTY_NUMBER.get(chartfile.difficulty, 3)),
    )


def slugify(s: str) -> str:
    s = remove_slashes(s)
    s = double_braces(s)
    return s


SLASHES = str.maketrans({"/": "", "\\": ""})


def remove_slashes(s: str) -> str:
    return s.translate(SLASHES)


BRACES = str.maketrans({"{": "{{", "}": "}}"})


def double_braces(s: str) -> str:
    return s.translate(BRACES)


class BetterStringFormatter(string.Formatter):
    """Enables the use of 'u' and 'l' suffixes in string format specifiers to
    convert the string to uppercase or lowercase
    Thanks stackoverflow ! https://stackoverflow.com/a/57570269/10768117"""

    def format_field(self, value: Any, format_spec: str) -> str:
        if isinstance(value, str):
            if format_spec.endswith("u"):
                value = value.upper()
                format_spec = format_spec[:-1]
            elif format_spec.endswith("l"):
                value = value.lower()
                format_spec = format_spec[:-1]
        return super().format(value, format_spec)



import warnings
from decimal import Decimal
from fractions import Fraction
from functools import singledispatch
from pathlib import Path
from typing import Any, List, Optional, Tuple, Union

import simplejson as json

from jubeatools import song
from jubeatools.formats import timemap
from jubeatools.formats.load_tools import FolderLoader, make_folder_loader
from jubeatools.utils import none_or

from . import schema as malody


def load_malody(path: Path, **kwargs: Any) -> song.Song:
    files = load_folder(path)
    charts = [load_malody_file(d) for d in files.values()]
    return song.Song.from_monochart_instances(*charts)


def load_file(path: Path) -> Any:
    with path.open() as f:
        return json.load(f, use_decimal=True)


load_folder: FolderLoader[Any] = make_folder_loader("*.mc", load_file)


def load_malody_file(raw_dict: dict) -> song.Song:
    file: malody.Chart = malody.CHART_SCHEMA.load(raw_dict)
    if file.meta.mode != malody.Mode.PAD:
        raise ValueError("This file is not a Malody Pad Chart (Malody's jubeat mode)")

    bgm = find_bgm(file.note)
    metadata = load_metadata(file.meta, bgm)
    time_map = load_timing_info(file.time, bgm)
    timing = time_map.convert_to_timing_info()
    chart = song.Chart(level=Decimal(0), timing=timing, notes=load_notes(file.note))
    dif = file.meta.version or song.Difficulty.EXTREME.value
    return song.Song(metadata=metadata, charts={dif: chart})


def find_bgm(events: List[malody.Event]) -> Optional[malody.Sound]:
    sounds = [e for e in events if isinstance(e, malody.Sound)]
    bgms = [s for s in sounds if s.type == malody.SoundType.BACKGROUND_MUSIC]
    if not bgms:
        return None

    if len(bgms) > 1:
        warnings.warn(
            "This file defines more than one background music, the first one "
            "will be used"
        )

    return min(bgms, key=lambda b: tuple_to_beats(b.beat))


def load_metadata(meta: malody.Metadata, bgm: Optional[malody.Sound]) -> song.Metadata:
    return song.Metadata(
        title=meta.song.title,
        artist=meta.song.artist,
        audio=none_or(lambda b: Path(b.sound), bgm),
        cover=none_or(Path, meta.background),
    )


def load_timing_info(
    bpm_changes: List[malody.BPMEvent], bgm: Optional[malody.Sound]
) -> timemap.TimeMap:
    if bgm is None:
        offset = timemap.SecondsAtBeat(seconds=Fraction(0), beats=Fraction(0))
    else:
        offset = timemap.SecondsAtBeat(
            seconds=-Fraction(bgm.offset) / 1000, beats=tuple_to_beats(bgm.beat)
        )
    return timemap.TimeMap.from_beats(
        events=[
            timemap.BPMAtBeat(beats=tuple_to_beats(b.beat), BPM=Fraction(b.bpm))
            for b in bpm_changes
        ],
        offset=offset,
    )


def load_notes(events: List[malody.Event]) -> List[Union[song.TapNote, song.LongNote]]:
    # filter out sound events
    notes: List[Union[malody.TapNote, malody.LongNote]] = [
        e for e in events if isinstance(e, (malody.TapNote, malody.LongNote))
    ]
    return [load_note(n) for n in notes]


@singledispatch
def load_note(
    n: Union[malody.TapNote, malody.LongNote]
) -> Union[song.TapNote, song.LongNote]:
    raise NotImplementedError(f"Unknown note type : {type(n)}")


@load_note.register
def load_tap_note(n: malody.TapNote) -> song.TapNote:
    return song.TapNote(
        time=tuple_to_beats(n.beat), position=song.NotePosition.from_index(n.index)
    )


@load_note.register
def load_long_note(n: malody.LongNote) -> song.LongNote:
    start = tuple_to_beats(n.beat)
    end = tuple_to_beats(n.endbeat)
    return song.LongNote(
        time=start,
        position=song.NotePosition.from_index(n.index),
        duration=end - start,
        tail_tip=song.NotePosition.from_index(n.endindex),
    )


def tuple_to_beats(b: Tuple[int, int, int]) -> song.BeatsTime:
    return b[0] + song.BeatsTime(b[1], b[2])



import textwrap
from dataclasses import astuple, dataclass
from decimal import Decimal
from itertools import product, zip_longest
from pathlib import Path
from typing import Any, Dict, Iterator, List, Mapping, Optional, Set, Tuple, Union

from parsimonious import Grammar, NodeVisitor, ParseError
from parsimonious.nodes import Node

from jubeatools.song import (
    BeatsTime,
    BPMEvent,
    Chart,
    LongNote,
    Metadata,
    NotePosition,
    Preview,
    SecondsTime,
    Song,
    TapNote,
    Timing,
)
from jubeatools.utils import none_or

from ..command import is_command, parse_command
from ..load_tools import (
    CIRCLE_FREE_TO_NOTE_SYMBOL,
    EMPTY_BEAT_SYMBOLS,
    JubeatAnalyserParser,
    UnfinishedLongNote,
    find_long_note_candidates,
    is_empty_line,
    load_folder,
    pick_correct_long_note_candidates,
)
from ..symbols import CIRCLE_FREE_SYMBOLS


@dataclass
class NoteCluster:
    string: str

    def dump(self) -> str:
        return self.string


@dataclass
class Stop:
    duration: int

    def dump(self) -> str:
        return f"[{self.duration}]"


@dataclass
class BPM:
    value: Decimal

    def dump(self) -> str:
        return f"({self.value})"


Event = Union[NoteCluster, Stop, BPM]


def dump_timing_part(events: List[Event]) -> str:
    return f"|{''.join(e.dump() for e in events)}|"


@dataclass
class RawMemo2ChartLine:
    position: str
    timing: Optional[List[Event]]

    def __str__(self) -> str:
        if self.timing:
            return f"{self.position} {dump_timing_part(self.timing)}"
        else:
            return self.position


@dataclass
class Memo2ChartLine:
    """timing part only contains notes"""

    position: str
    timing: Optional[List[str]]

    @property
    def duration(self) -> BeatsTime:
        if self.timing:
            return BeatsTime(1)
        else:
            return BeatsTime(0)


memo2_chart_line_grammar = Grammar(
    r"""
    line            = ws position_part ws (timing_part ws)? comment?
    position_part   = ~r"[^*#:|/\s]{4,8}"
    timing_part     = "|" event+ "|"
    event           = stop / bpm / notes
    stop            = "[" pos_integer "]"
    pos_integer     = ~r"\d+"
    bpm             = "(" float ")"
    float           = ~r"\d+(\.\d+)?"
    notes           = ~r"[^*#:\(\[|/\s]+"
    ws              = ~r"[\t \u3000]*" # U+03000 : IDEOGRAPHIC SPACE
    comment         = ~r"//.*"
"""
)


class Memo2ChartLineVisitor(NodeVisitor):
    def __init__(self) -> None:
        super().__init__()
        self.pos_part: Optional[str] = None
        self.time_part: List[Union[NoteCluster, BPM, Stop]] = []

    def visit_line(self, node: Node, visited_children: list) -> RawMemo2ChartLine:
        if self.pos_part is None:
            raise ValueError("Line has no position part")

        return RawMemo2ChartLine(self.pos_part, self.time_part or None)

    def visit_position_part(self, node: Node, visited_children: list) -> None:
        self.pos_part = node.text

    def visit_stop(self, node: Node, visited_children: list) -> None:
        _, duration, _ = node.children
        self.time_part.append(Stop(int(duration.text)))

    def visit_bpm(self, node: Node, visited_children: list) -> None:
        _, value, _ = node.children
        self.time_part.append(BPM(Decimal(value.text)))

    def visit_notes(self, node: Node, visited_children: list) -> None:
        self.time_part.append(NoteCluster(node.text))

    def generic_visit(self, node: Node, visited_children: list) -> None:
        ...


def is_memo2_chart_line(line: str) -> bool:
    try:
        memo2_chart_line_grammar.parse(line)
    except ParseError:
        return False
    else:
        return True


def parse_memo2_chart_line(line: str) -> RawMemo2ChartLine:
    return Memo2ChartLineVisitor().visit(memo2_chart_line_grammar.parse(line))  # type: ignore


@dataclass
class Memo2Frame:
    position_part: List[List[str]]
    timing_part: List[List[str]]

    @property
    def duration(self) -> BeatsTime:
        return BeatsTime(len(self.timing_part))

    def __str__(self) -> str:
        res = []
        for pos, time in zip_longest(self.position_part, self.timing_part):
            line = [f"{''.join(pos)}"]
            if time is not None:
                line += [f"|{''.join(time)}|"]
            res += [" ".join(line)]
        return "\n".join(res)


CHARACTER_WIDTH_HELP = textwrap.dedent(
    """
    Make sure the file is :
      - truly encoded as Shift-JIS
      - not missusing half-width characters instead of full-width ones
      - only using half-width characters in pairs
    """
).strip()


class Memo2Parser(JubeatAnalyserParser):

    FORMAT_TAG = "#memo2"

    def __init__(self) -> None:
        super().__init__()
        self.current_chart_lines: List[Memo2ChartLine] = []
        self.frames: List[Memo2Frame] = []

    def do_b(self, value: str) -> None:
        if Decimal(value) != 4:
            raise RuntimeError(
                "Beat command (b=...) found with a value other that 4, this is "
                "unexpected in #memo2 files. If you are sure the file is not "
                "just actually a #memo1 or #memo file with the wrong format "
                "tag, you're welcome to report this error as a bug"
            )

    def do_t(self, value: str) -> None:
        self.timing_events.append(BPMEvent(self._current_beat(), BPM=Decimal(value)))

    def do_r(self, value: str) -> None:
        if self.frames:
            raise RuntimeError(
                "offset increase command (r=...)  found outside of the file "
                "header, this is not supported by jubeatools"
            )
        else:
            super().do_r(value)

    def do_boogie(self) -> None:
        self.do_bpp(1)

    def do_memo2(self) -> None:
        ...

    def do_bpp(self, value: Union[int, str]) -> None:
        if self.frames:
            raise RuntimeError("jubeatools does not handle changes of #bpp halfway")
        else:
            self._do_bpp(value)

    def append_chart_line(self, raw_line: RawMemo2ChartLine) -> None:
        position_length = len(
            raw_line.position.encode("shift-jis-2004", errors="surrogateescape")
        )
        expected_position_length = 4 * self.bytes_per_panel
        if position_length != expected_position_length:
            message = [
                textwrap.dedent(
                    f"""
                The position part of this chart line :

                  {raw_line.position}

                should be {expected_position_length} bytes long (4 × #bpp value) but it is {position_length}
                """
                ).strip()
            ]
            message.append(CHARACTER_WIDTH_HELP)
            raise SyntaxError("\n".join(message))

        if raw_line.timing is not None and self.bytes_per_panel == 2:
            for note_cluster in (
                elem for elem in raw_line.timing if isinstance(elem, NoteCluster)
            ):
                length = len(
                    note_cluster.string.encode(
                        "shift-jis-2004", errors="surrogateescape"
                    )
                )
                if length % 2 != 0:
                    timing_part = dump_timing_part(raw_line.timing)
                    message = [
                        textwrap.dedent(
                            f"""
                        The timing part of this chart line :
                        
                          {timing_part}
                        
                        contains the following {length} bytes long series of notes :

                          {note_cluster.string}

                        Series of notes of odd byte lengths should not happen in files that use #bpp=2
                        """
                        ).strip()
                    ]
                    message.append(CHARACTER_WIDTH_HELP)
                    raise SyntaxError("\n".join(message))

        if not raw_line.timing:
            line = Memo2ChartLine(raw_line.position, None)
        else:
            # split notes
            bar: List[Union[str, Stop, BPM]] = []
            for raw_event in raw_line.timing:
                if isinstance(raw_event, NoteCluster):
                    bar.extend(self._split_chart_line(raw_event.string))
                else:
                    bar.append(raw_event)
            # extract timing info
            bar_length = sum(1 for e in bar if isinstance(e, str))
            symbol_duration = BeatsTime(1, bar_length)
            in_bar_beat = BeatsTime(0)
            for event in bar:
                if isinstance(event, str):
                    in_bar_beat += symbol_duration
                elif isinstance(event, BPM):
                    self.timing_events.append(
                        BPMEvent(
                            time=self._current_beat() + in_bar_beat, BPM=event.value
                        )
                    )
                elif isinstance(event, Stop):
                    time = self._current_beat() + in_bar_beat
                    if time != 0:
                        raise ValueError(
                            "Chart contains a pause that's not happening at the "
                            "very first beat, these are not supported by jubeatools"
                        )
                    self.offset += event.duration

            bar_notes = [e for e in bar if isinstance(e, str)]
            line = Memo2ChartLine(raw_line.position, bar_notes)

        self.current_chart_lines.append(line)
        if len(self.current_chart_lines) == 4:
            self._push_frame()

    def _current_beat(self) -> BeatsTime:
        return self._frames_duration() + self._lines_duration()

    def _frames_duration(self) -> BeatsTime:
        return sum((frame.duration for frame in self.frames), start=BeatsTime(0))

    def _lines_duration(self) -> BeatsTime:
        return sum(
            (line.duration for line in self.current_chart_lines), start=BeatsTime(0)
        )

    def _push_frame(self) -> None:
        position_part = [
            self._split_chart_line(memo_line.position)
            for memo_line in self.current_chart_lines
        ]
        timing_part = [
            memo_line.timing
            for memo_line in self.current_chart_lines
            if memo_line.timing is not None
        ]
        frame = Memo2Frame(position_part, timing_part)
        self.frames.append(frame)
        self.current_chart_lines = []

    def finish_last_few_notes(self) -> None:
        """Call this once when the end of the file is reached,
        flushes the chart line and chart frame buffers to create the last chart
        section"""
        if self.current_chart_lines:
            if len(self.current_chart_lines) != 4:
                raise SyntaxError(
                    f"Unfinished chart frame when flushing : \n"
                    f"{self.current_chart_lines}"
                )
            self._push_frame()

    def load_line(self, raw_line: str) -> None:
        line = raw_line.strip()
        self.raise_if_separator(line, self.FORMAT_TAG)
        if is_command(line):
            command, value = parse_command(line)
            self.handle_command(command, value)
        elif is_empty_line(line) or self.is_short_line(line):
            return
        elif is_memo2_chart_line(line):
            memo_chart_line = parse_memo2_chart_line(line)
            self.append_chart_line(memo_chart_line)
        else:
            raise SyntaxError(f"not a valid memo2 file line : {line}")

    def notes(self) -> Iterator[Union[TapNote, LongNote]]:
        if self.hold_by_arrow:
            yield from self._iter_notes()
        else:
            yield from self._iter_notes_without_longs()

    def _iter_frames(
        self,
    ) -> Iterator[Tuple[Mapping[str, BeatsTime], Memo2Frame]]:
        """iterate over tuples of (currently_defined_symbols, frame)"""
        local_symbols = {}
        frame_starting_beat = BeatsTime(0)
        for i, frame in enumerate(self.frames):
            if frame.timing_part:
                frame_starting_beat = sum(
                    (f.duration for f in self.frames[:i]), start=BeatsTime(0)
                )
                local_symbols = {
                    symbol: frame_starting_beat
                    + bar_index
                    + BeatsTime(symbol_index, len(bar))
                    for bar_index, bar in enumerate(frame.timing_part)
                    for symbol_index, symbol in enumerate(bar)
                    if symbol not in EMPTY_BEAT_SYMBOLS
                }
            yield local_symbols, frame

    def _iter_notes(self) -> Iterator[Union[TapNote, LongNote]]:
        unfinished_longs: Dict[NotePosition, UnfinishedLongNote] = {}
        for currently_defined_symbols, frame in self._iter_frames():
            should_skip: Set[NotePosition] = set()
            # 1/3 : look for ends to unfinished long notes
            for pos, unfinished_long in unfinished_longs.items():
                x, y = astuple(pos)
                symbol = frame.position_part[y][x]
                if self.circle_free and symbol in CIRCLE_FREE_SYMBOLS:
                    circled_symbol = CIRCLE_FREE_TO_NOTE_SYMBOL[symbol]
                    try:
                        symbol_time = currently_defined_symbols[circled_symbol]
                    except KeyError:
                        raise SyntaxError(
                            "Chart section positional part constains the circle free "
                            f"symbol '{symbol}' but the associated circled symbol "
                            f"'{circled_symbol}' could not be found in the timing part:\n"
                            f"{frame}"
                        )
                else:
                    try:
                        symbol_time = currently_defined_symbols[symbol]
                    except KeyError:
                        continue

                should_skip.add(pos)
                yield unfinished_long.ends_at(symbol_time)

            unfinished_longs = {
                k: unfinished_longs[k] for k in unfinished_longs.keys() - should_skip
            }

            # 2/3 : look for new long notes starting on this bloc
            arrow_to_note_candidates = find_long_note_candidates(
                frame.position_part, currently_defined_symbols.keys(), should_skip
            )
            if arrow_to_note_candidates:
                solution = pick_correct_long_note_candidates(
                    arrow_to_note_candidates,
                    frame.position_part,
                )
                for arrow_pos, note_pos in solution.items():
                    should_skip.add(arrow_pos)
                    should_skip.add(note_pos)
                    symbol = frame.position_part[note_pos.y][note_pos.x]
                    symbol_time = currently_defined_symbols[symbol]
                    unfinished_longs[note_pos] = UnfinishedLongNote(
                        time=symbol_time, position=note_pos, tail_tip=arrow_pos
                    )

            # 3/3 : find regular notes
            for y, x in product(range(4), range(4)):
                position = NotePosition(x, y)
                if position in should_skip:
                    continue
                symbol = frame.position_part[y][x]
                try:
                    symbol_time = currently_defined_symbols[symbol]
                except KeyError:
                    continue
                yield TapNote(symbol_time, position)

    def _iter_notes_without_longs(self) -> Iterator[TapNote]:
        for currently_defined_symbols, frame in self._iter_frames():
            # cross compare symbols with the position information
            for y, x in product(range(4), range(4)):
                symbol = frame.position_part[y][x]
                try:
                    symbol_time = currently_defined_symbols[symbol]
                except KeyError:
                    continue
                position = NotePosition(x, y)
                yield TapNote(symbol_time, position)


def _load_memo2_file(lines: List[str]) -> Song:
    parser = Memo2Parser()
    for i, raw_line in enumerate(lines, start=1):
        try:
            parser.load_line(raw_line)
        except Exception as e:
            raise SyntaxError(f"On line {i}\n{e}")

    parser.finish_last_few_notes()
    metadata = Metadata(
        title=parser.title,
        artist=parser.artist,
        audio=none_or(Path, parser.music),
        cover=none_or(Path, parser.jacket),
    )
    if parser.preview_start is not None:
        metadata.preview = Preview(
            start=SecondsTime(parser.preview_start) / 1000, length=SecondsTime(10)
        )

    timing = Timing(
        events=parser.timing_events,
        beat_zero_offset=SecondsTime(parser.offset or 0) / 1000,
    )
    charts = {
        parser.difficulty
        or "EXT": Chart(
            level=Decimal(parser.level),
            timing=timing,
            notes=sorted(parser.notes(), key=lambda n: (n.time, n.position)),
        )
    }
    return Song(metadata=metadata, charts=charts)


def load_memo2(path: Path, **kwargs: Any) -> Song:
    files = load_folder(path)
    charts = [_load_memo2_file(lines) for _, lines in files.items()]
    return Song.from_monochart_instances(*charts)



"""Collection of tools related to dumping to jubeat analyser formats"""

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from decimal import Decimal
from fractions import Fraction
from itertools import chain
from pathlib import Path
from typing import (
    Any,
    Callable,
    DefaultDict,
    Dict,
    Iterator,
    List,
    Mapping,
    TypedDict,
    TypeVar,
    Union,
)

from more_itertools import windowed
from sortedcontainers import SortedDict, SortedKeyList

from jubeatools.formats.dump_tools import (
    DIFFICULTY_NUMBER,
    make_dumper_from_chart_file_dumper,
)
from jubeatools.formats.filetypes import ChartFile
from jubeatools.formats.typing import Dumper
from jubeatools.song import (
    BeatsTime,
    Chart,
    Direction,
    LongNote,
    Metadata,
    NotePosition,
    Song,
    TapNote,
    Timing,
)
from jubeatools.utils import fraction_to_decimal

from .command import dump_command
from .symbols import CIRCLE_FREE_SYMBOLS, NOTE_SYMBOLS
from .typing import JubeatAnalyserChartDumper

COMMAND_ORDER = [
    "b",
    "t",
    "m",
    "o",
    "r",
    "title",
    "artist",
    "lev",
    "dif",
    "jacket",
    "prevpos",
    "holdbyarrow",
    "circlefree",
]

BEATS_TIME_TO_SYMBOL = {
    BeatsTime(1, 4) * index: symbol for index, symbol in enumerate(NOTE_SYMBOLS)
}

BEATS_TIME_TO_CIRCLE_FREE = {
    BeatsTime(1, 4) * index: symbol for index, symbol in enumerate(CIRCLE_FREE_SYMBOLS)
}

NOTE_TO_CIRCLE_FREE_SYMBOL = dict(zip(NOTE_SYMBOLS, CIRCLE_FREE_SYMBOLS))

DIRECTION_TO_ARROW = {
    Direction.RIGHT: "＞",  # U+FF1E : FULLWIDTH GREATER-THAN SIGN
    Direction.LEFT: "＜",  # U+FF1C : FULLWIDTH LESS-THAN SIGN
    Direction.DOWN: "∨",  # U+2228 : LOGICAL OR
    Direction.UP: "∧",  # U+2227 : LOGICAL AND
}

# do NOT use the regular vertical bar, it will clash with the timing portion
DIRECTION_TO_LINE = {
    Direction.RIGHT: "―",  # U+2015 : HORIZONTAL BAR
    Direction.LEFT: "―",
    Direction.UP: "｜",  # U+FF5C : FULLWIDTH VERTICAL LINE
    Direction.DOWN: "｜",
}

# I put a FUCKTON of extra characters just in case some insane chart uses
# loads of unusual beat divisions.
# /!\ The Vs are left out on purpose since they would be mistaken for
# long note arrows
DEFAULT_EXTRA_SYMBOLS = (
    "ＡＢＣＤＥＦＧＨＩＪＫＬＭＮＯＰＱＲＳＴＵＷＸＹＺ"
    "ａｂｃｄｅｆｇｈｉｊｋｌｍｎｏｐｑｒｓｔｕｗｘｙｚ"
    "あいうえおかきくけこさしすせそたちつてとなにぬねのはひふへほまみむめもやゆよらりるれろわをん"
    "アイウエオカキクケコサシスセソタチツテトナニヌネノハヒフヘホマミムメモヤユヨラリルレロワヲン"
)


@dataclass(frozen=True)
class LongNoteEnd:
    time: BeatsTime
    position: NotePosition


K = TypeVar("K")
V = TypeVar("V")


class SortedDefaultDict(SortedDict, DefaultDict[K, V]):

    """Custom SortedDict that also acts as a defaultdict,
    passes the key to the value factory"""

    def __init__(self, factory: Callable[[K], V], *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.__factory__ = factory

    def add_key(self, key: K) -> None:
        if key not in self:
            value = self.__factory__(key)
            self.__setitem__(key, value)

    def __missing__(self, key: K) -> V:
        value = self.__factory__(key)
        self.__setitem__(key, value)
        return value


class Commands(TypedDict, total=False):
    b: Union[int, Fraction, Decimal]
    t: Union[int, Fraction, Decimal]
    m: Union[str, Path]
    o: int
    r: int
    title: str
    artist: str
    lev: Union[int, Decimal]
    dif: int
    jacket: Union[str, Path]
    prevpos: int
    holdbyarrow: int
    circlefree: int
    memo: None
    memo1: None
    memo2: None
    boogie: None
    pw: int
    ph: int
    bpp: int


# Here we split dataclass and ABC stuff since mypy curently can't handle both
# at once on a single class definition
@dataclass
class _JubeatAnalyerDumpedSection:
    current_beat: BeatsTime
    length: BeatsTime = BeatsTime(4)
    commands: Commands = field(default_factory=Commands)  # type: ignore
    symbol_definitions: Dict[BeatsTime, str] = field(default_factory=dict)
    symbols: Dict[BeatsTime, str] = field(default_factory=dict)
    notes: List[Union[TapNote, LongNote, LongNoteEnd]] = field(default_factory=list)


class JubeatAnalyserDumpedSection(_JubeatAnalyerDumpedSection, ABC):
    def _dump_commands(self) -> Iterator[str]:
        keys = chain(COMMAND_ORDER, self.commands.keys() - set(COMMAND_ORDER))
        for key in keys:
            if key in self.commands:
                yield dump_command(key, self.commands[key])  # type: ignore

    def _dump_symbol_definitions(self) -> Iterator[str]:
        for time, symbol in self.symbol_definitions.items():
            decimal_time = fraction_to_decimal(time)
            yield f"*{symbol}:{decimal_time:.6}"

    @abstractmethod
    def _dump_notes(self, circle_free: bool) -> Iterator[str]:
        ...

    @abstractmethod
    def render(self, circle_free: bool) -> str:
        ...


def create_sections_from_chart(
    section_factory: Callable[[BeatsTime], JubeatAnalyserDumpedSection],
    chart: Chart,
    difficulty: str,
    timing: Timing,
    metadata: Metadata,
    circle_free: bool,
) -> Mapping[BeatsTime, JubeatAnalyserDumpedSection]:
    sections = SortedDefaultDict(section_factory)

    timing_events = sorted(timing.events, key=lambda e: e.time)
    notes = SortedKeyList(set(chart.notes), key=lambda n: n.time)

    for note in chart.notes:
        if isinstance(note, LongNote):
            notes.add(LongNoteEnd(note.time + note.duration, note.position))

    all_events = SortedKeyList(timing_events + notes, key=lambda n: n.time)
    last_event = all_events[-1]
    last_measure = last_event.time // 4
    for i in range(last_measure + 1):
        beat = BeatsTime(4) * i
        sections.add_key(beat)

    header = sections[BeatsTime(0)].commands
    header["o"] = int(timing.beat_zero_offset * 1000)
    if chart.level is not None:
        header["lev"] = Decimal(chart.level)
    header["dif"] = DIFFICULTY_NUMBER.get(difficulty, 3)
    if metadata.audio is not None:
        header["m"] = metadata.audio
    if metadata.title is not None:
        header["title"] = metadata.title
    if metadata.artist is not None:
        header["artist"] = metadata.artist
    if metadata.cover is not None:
        header["jacket"] = metadata.cover
    if metadata.preview is not None:
        header["prevpos"] = int(metadata.preview.start * 1000)

    if any(isinstance(note, LongNote) for note in chart.notes):
        header["holdbyarrow"] = 1

    if circle_free:
        header["circlefree"] = 1

    # Potentially create sub-sections for bpm changes
    for event in timing_events:
        sections[event.time].commands["t"] = event.BPM

    # First, Set every single b=… value
    for key, next_key in windowed(chain(sections.keys(), [None]), 2):
        if key is None:
            continue
        elif next_key is None:
            length = BeatsTime(4)
        else:
            length = next_key - key

        sections[key].commands["b"] = length
        sections[key].length = length

    # Then, trim all the redundant b=…
    last_b: Union[int, Fraction, Decimal] = 4
    for section in sections.values():
        current_b = section.commands["b"]
        if current_b == last_b:
            del section.commands["b"]
        else:
            last_b = current_b

    # Fill sections with notes
    for key, next_key in windowed(chain(sections.keys(), [None]), 2):
        assert key is not None
        sections[key].notes = list(
            notes.irange_key(min_key=key, max_key=next_key, inclusive=(True, False))
        )

    return sections


def make_full_dumper_from_jubeat_analyser_chart_dumper(
    chart_dumper: JubeatAnalyserChartDumper,
) -> Dumper:
    """Factory function to create a fully fledged song dumper from
    the internal chart dumper of jubeat analyser formats"""

    def song_dumper(
        song: Song, *, circle_free: bool = False, **kwargs: Any
    ) -> List[ChartFile]:
        files: List[ChartFile] = []
        for difficulty, chart, timing in song.iter_charts_with_applicable_timing():
            chart_file = chart_dumper(
                difficulty,
                chart,
                song.metadata,
                timing,
                circle_free,
            )
            file_bytes = chart_file.getvalue().encode(
                "shift-jis-2004", errors="surrogateescape"
            )
            files.append(ChartFile(file_bytes, song, difficulty, chart))

        return files

    return make_dumper_from_chart_file_dumper(
        internal_dumper=song_dumper,
        file_name_template=Path("{title} {difficulty_number}.txt"),
    )



from pathlib import Path
from typing import Any, Iterator, List

from jubeatools import song
from jubeatools.formats.load_tools import make_folder_loader

from ..commons import Command, Event
from ..load_tools import make_chart_from_events


def load_eve(path: Path, *, beat_snap: int = 240, **kwargs: Any) -> song.Song:
    files = load_folder(path)
    charts = [_load_eve(l, p, beat_snap=beat_snap) for p, l in files.items()]
    return song.Song.from_monochart_instances(*charts)


def load_file(path: Path) -> List[str]:
    return path.read_text(encoding="ascii").split("\n")


load_folder = make_folder_loader("*.eve", load_file)


def _load_eve(lines: List[str], file_path: Path, *, beat_snap: int = 240) -> song.Song:
    chart = make_chart_from_events(iter_events(lines), beat_snap=beat_snap)
    dif = guess_difficulty(file_path.stem)
    return song.Song(metadata=song.Metadata(), charts={dif.value: chart})


def iter_events(lines: List[str]) -> Iterator[Event]:
    for i, raw_line in enumerate(lines, start=1):
        line = raw_line.strip()
        if not line:
            continue

        try:
            yield parse_event(line)
        except ValueError as e:
            raise ValueError(f"Error on line {i} : {e}")


def parse_event(line: str) -> Event:
    columns = line.split(",")
    if len(columns) != 3:
        raise ValueError(f"Expected 3 comma-separated values but found {len(columns)}")

    raw_tick, raw_command, raw_value = map(str.strip, columns)
    try:
        tick = int(raw_tick)
    except ValueError:
        raise ValueError(
            f"The first column should contain an integer but {raw_tick!r} was "
            f"found, which python could not understand as an integer"
        )

    try:
        command = Command[raw_command]
    except KeyError:
        raise ValueError(
            f"The second column should contain one of "
            f"{list(Command.__members__)}, but {raw_command!r} was found"
        )

    try:
        value = int(raw_value)
    except ValueError:
        raise ValueError(
            f"The third column should contain an integer but {raw_tick!r} was "
            f"found, which python could not understand as an integer"
        )

    return Event(tick, command, value)


def guess_difficulty(filename: str) -> song.Difficulty:
    try:
        return song.Difficulty(filename.upper())
    except ValueError:
        return song.Difficulty.EXTREME



from pathlib import Path
from typing import List

from jubeatools import song
from jubeatools.formats.dump_tools import make_dumper_from_chart_file_dumper
from jubeatools.formats.filetypes import ChartFile

from ..dump_tools import make_events_from_chart


def _dump_eve(song: song.Song, **kwargs: dict) -> List[ChartFile]:
    res = []
    for dif, chart, timing, hakus in song.iter_charts():
        events = make_events_from_chart(chart.notes, timing, hakus)
        chart_text = "\n".join(e.dump() for e in events)
        chart_bytes = chart_text.encode("ascii")
        res.append(ChartFile(chart_bytes, song, dif, chart))

    return res


dump_eve = make_dumper_from_chart_file_dumper(
    internal_dumper=_dump_eve, file_name_template=Path("{difficulty:l}.eve")
)



from fractions import Fraction

from hypothesis import given

from jubeatools import song
from jubeatools.formats.timemap import TimeMap
from jubeatools.testutils import strategies as jbst
from jubeatools.utils import group_by


@given(jbst.timing_info(with_bpm_changes=True), jbst.beat_time())
def test_that_seconds_at_beat_works_like_the_naive_approach(
    timing: song.Timing, beat: song.BeatsTime
) -> None:
    time_map = TimeMap.from_timing(timing)
    expected = naive_approach(timing, beat)
    actual = time_map.fractional_seconds_at(beat)
    assert actual == expected


def naive_approach(beats: song.Timing, beat: song.BeatsTime) -> Fraction:
    if beat < 0:
        raise ValueError("Can't compute seconds at negative beat")

    if not beats.events:
        raise ValueError("No BPM defined")

    grouped_by_time = group_by(beats.events, key=lambda e: e.time)
    for time, events in grouped_by_time.items():
        if len(events) > 1:
            raise ValueError(f"Multiple BPMs defined on beat {time} : {events}")

    sorted_events = sorted(beats.events, key=lambda e: e.time)
    first_event = sorted_events[0]
    if first_event.time != song.BeatsTime(0):
        raise ValueError("First BPM event is not on beat zero")

    if beat > sorted_events[-1].time:
        events_before = sorted_events
    else:
        last_index = next(i for i, e in enumerate(sorted_events) if e.time >= beat)
        events_before = sorted_events[:last_index]
    total_seconds = Fraction(0)
    current_beat = beat
    for event in reversed(events_before):
        beats_since_previous = current_beat - event.time
        seconds_since_previous = (60 * beats_since_previous) / Fraction(event.BPM)
        total_seconds += seconds_since_previous
        current_beat = event.time

    total_seconds = total_seconds + Fraction(beats.beat_zero_offset)
    return total_seconds



from enum import Enum


class Format(str, Enum):
    EVE = "eve"
    JBSQ = "jbsq"
    MALODY = "malody"
    MEMON_LEGACY = "memon:legacy"
    MEMON_0_1_0 = "memon:v0.1.0"
    MEMON_0_2_0 = "memon:v0.2.0"
    MEMON_0_3_0 = "memon:v0.3.0"
    MEMON_1_0_0 = "memon:v1.0.0"
    MONO_COLUMN = "mono-column"
    MEMO = "memo"
    MEMO_1 = "memo1"
    MEMO_2 = "memo2"



from decimal import Decimal
from fractions import Fraction

from ..load_tools import round_beats


def test_fraction_recovery_after_rounding_to_three_decimals() -> None:
    for numerator in range(240):
        fraction = Fraction(numerator, 240)
        decimal = numerator / Decimal(240)
        rounded = round(decimal, 3)
        text_form = str(rounded)
        re_parsed_decimal = Decimal(text_form)
        result = round_beats(re_parsed_decimal)
        assert fraction == result



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['jubeatools',
 'jubeatools.cli',
 'jubeatools.cli.tests',
 'jubeatools.cli.tests.data',
 'jubeatools.formats',
 'jubeatools.formats.jubeat_analyser',
 'jubeatools.formats.jubeat_analyser.memo',
 'jubeatools.formats.jubeat_analyser.memo1',
 'jubeatools.formats.jubeat_analyser.memo2',
 'jubeatools.formats.jubeat_analyser.mono_column',
 'jubeatools.formats.jubeat_analyser.tests',
 'jubeatools.formats.jubeat_analyser.tests.data',
 'jubeatools.formats.jubeat_analyser.tests.memo',
 'jubeatools.formats.jubeat_analyser.tests.memo1',
 'jubeatools.formats.jubeat_analyser.tests.memo2',
 'jubeatools.formats.jubeat_analyser.tests.mono_column',
 'jubeatools.formats.konami',
 'jubeatools.formats.konami.eve',
 'jubeatools.formats.konami.eve.tests',
 'jubeatools.formats.konami.jbsq',
 'jubeatools.formats.konami.jbsq.tests',
 'jubeatools.formats.malody',
 'jubeatools.formats.malody.tests',
 'jubeatools.formats.malody.tests.data',
 'jubeatools.formats.memon',
 'jubeatools.formats.memon.v0',
 'jubeatools.formats.memon.v1',
 'jubeatools.formats.memon.v1.tests',
 'jubeatools.formats.tests',
 'jubeatools.testutils']

package_data = \
{'': ['*'], 'jubeatools.cli.tests.data': ['memon_merge/*']}

install_requires = \
['click>=8.0.3,<9.0.0',
 'construct-typing>=0.5.3,<0.6.0',
 'construct>=2.10,<3.0',
 'marshmallow-dataclass[enum,union]>=8.5.3,<9.0.0',
 'marshmallow>=3.6.0,<4.0.0',
 'more-itertools>=8.4.0,<9.0.0',
 'parsimonious>=0.10.0,<0.11.0',
 'python-constraint>=1.4.0,<2.0.0',
 'simplejson>=3.17.0,<4.0.0',
 'sortedcontainers>=2.3.0,<3.0.0',
 'typing-inspect==0.7.1']

entry_points = \
{'console_scripts': ['jubeatools = jubeatools.cli.cli:convert']}

setup_kwargs = {
    'name': 'jubeatools',
    'version': '2.0.3',
    'description': 'A toolbox for jubeat file formats',
    'long_description': '# Jubeatools\n\n[![docs status badge](https://readthedocs.org/projects/jubeatools/badge/)](https://jubeatools.readthedocs.io)\n\nA toolbox to convert between jubeat file formats\n\n## How to install\n```sh\npip install jubeatools\n```\n\nYou need Python 3.9 or greater\n\nMore details in [the documentation](https://jubeatools.readthedocs.io)\n\n## How to use in a command line\n```sh\njubeatools ${source} ${destination} -f ${output format} (... format specific options)\n```\n\nAgain, more details in [the documentation](https://jubeatools.readthedocs.io)\n\n## Which formats are supported\n|                 |                      | input | output |\n|-----------------|----------------------|:-----:|:------:|\n| memon           | v1.0.0               | ✔️     | ✔️      |\n|                 | v0.3.0               | ✔️     | ✔️      |\n|                 | v0.2.0               | ✔️     | ✔️      |\n|                 | v0.1.0               | ✔️     | ✔️      |\n|                 | legacy               | ✔️     | ✔️      |\n| jubeat analyser | #memo2               | ✔️     | ✔️      |\n|                 | #memo1               | ✔️     | ✔️      |\n|                 | #memo                | ✔️     | ✔️      |\n|                 | mono-column (1列形式) | ✔️     | ✔️      |\n| jubeat (arcade) | .eve                 | ✔️     | ✔️      |\n| jubeat plus     | .jbsq                | ✔️     | ✔️      |\n| malody          | .mc (Pad Mode)       | ✔️     | ✔️      |\n',
    'author': 'Stepland',
    'author_email': '10530295-Buggyroom@users.noreply.gitlab.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/square-game-liberation-front/jubeatools',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from io import StringIO
from typing import Protocol

from jubeatools.song import Chart, Metadata, Timing


class JubeatAnalyserChartDumper(Protocol):
    """Internal Dumper for jubeat analyser formats"""

    def __call__(
        self,
        difficulty: str,
        chart: Chart,
        metadata: Metadata,
        timing: Timing,
        circle_free: bool = False,
    ) -> StringIO:
        ...



from decimal import Decimal
from fractions import Fraction

from jubeatools.song import *

data = Song(
    metadata=Metadata(
        title=None, artist=None, audio=None, cover=None, preview=None, preview_file=None
    ),
    charts={
        "BSC": Chart(
            level=Decimal("0"),
            timing=Timing(
                events=(BPMEvent(time=Fraction(0, 1), BPM=Decimal("288.11")),),
                beat_zero_offset=Decimal("0.04"),
            ),
            hakus=set(),
            notes=[],
        )
    },
    common_timing=None,
    common_hakus=None,
)



import tempfile
from contextlib import contextmanager
from pathlib import Path
from typing import Callable, ContextManager, Iterator, Optional

from hypothesis import note

from jubeatools import song
from jubeatools.formats import DUMPERS, LOADERS, Format
from jubeatools.formats.guess import guess_format


@contextmanager
def open_temp_dir() -> Iterator[Path]:
    with tempfile.TemporaryDirectory() as temp_dir:
        yield Path(temp_dir)


def dump_and_load_then_compare(
    format_: Format,
    song: song.Song,
    bytes_decoder: Callable[[bytes], str] = lambda b: b.decode("utf-8"),
    temp_path: Callable[[], ContextManager[Path]] = open_temp_dir,
    load_options: Optional[dict] = None,
    dump_options: Optional[dict] = None,
) -> None:
    recovered_song = dump_and_load(
        format_, song, bytes_decoder, temp_path, load_options, dump_options
    )
    assert recovered_song == song


def dump_and_load(
    format_: Format,
    song: song.Song,
    bytes_decoder: Callable[[bytes], str] = lambda b: b.decode("utf-8"),
    temp_path: Callable[[], ContextManager[Path]] = open_temp_dir,
    load_options: Optional[dict] = None,
    dump_options: Optional[dict] = None,
) -> song.Song:
    load_options = load_options or {}
    dump_options = dump_options or {}
    loader = LOADERS[format_]
    dumper = DUMPERS[format_]
    song.minimize_timings()
    song.minimize_hakus()
    with temp_path() as folder_path:
        files = dumper(song, folder_path, **dump_options)
        for file_path, bytes_ in files.items():
            file_path.write_bytes(bytes_)
            note(f"Wrote to {file_path} :\n{bytes_decoder(bytes_)}")
            assert guess_format(file_path) == format_
        recovered_song = loader(folder_path, **load_options)
        recovered_song.minimize_timings()
        recovered_song.minimize_hakus()
        return recovered_song



"""
Useful things to parse and dump the jubeat analyser command format

Known simple commands :
  - b=<decimal>   : beats per measure (4 by default)
  - m="<path>"    : music file path
  - o=<int>       : offset in ms (100 by default)
  - r=<int>       : ? increase the offset ? (in ms) (not supported, couldn't find any examples, wtf is it for ?)
  - t=<decimal>   : tempo
  
Known hash commands :
  - #memo             # youbeat-like format but a bar division always means a 1/4 note (amongst other quirks)
  - #memo1            # youbeat-like but without bpm changes in the bar
  - #memo2            # youbeat-like but double-width
  - #boogie           # youbeat-like
  - #pw=<int>         # number of panels horizontally (4 by default)
  - #ph=<int>         # number of panels vertically (4 by default)
  - #lev=<int>        # chart level (typically 1 to 10)
  - #dif={1, 2, 3}    # 1: BSC, 2: ADV, 3: EXT
  - #title="<str>"    # music title
  - #artist="<str>"   # artist's name
  - #jacket="<path>"  # music cover art path
  - #prevpos=<int>    # preview start (in ms)
  - #bpp              # bytes per panel (2 by default)
"""

from decimal import ROUND_HALF_DOWN, Decimal
from fractions import Fraction
from functools import singledispatch
from pathlib import Path
from typing import Any, List, Optional, Tuple, Union

from parsimonious import Grammar, NodeVisitor, ParseError
from parsimonious.nodes import Node

from jubeatools.utils import fraction_to_decimal

command_grammar = Grammar(
    r"""
    line            = ws command ws comment?
    command         = hash_command / short_command
    hash_command    = "#" key equals_value?
    short_command   = letter equals_value
    letter          = ~r"\w"
    key             = ~r"\w+"
    equals_value    = ws "=" ws value
    value           = value_in_quotes / number
    value_in_quotes = '"' quoted_value '"'
    quoted_value    = ~r"([^\"\\]|\\\"|\\\\)*"
    number          = ~r"-?\d+(\.\d+)?"
    ws              = ~r"[\t \u3000]*"
    comment         = ~r"//.*"
    """
)


class CommandVisitor(NodeVisitor):

    """Returns a (key, value) tuple"""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.key: Optional[str] = None
        self.value: Optional[str] = None

    def visit_line(
        self, node: Node, visited_children: List[Node]
    ) -> Tuple[str, Optional[str]]:
        if self.key is None:
            raise ValueError("No key found after parsing command")
        return self.key, self.value

    def visit_hash_command(self, node: Node, visited_children: List[Node]) -> None:
        _, key, _ = node.children
        self.key = key.text

    def visit_short_command(self, node: Node, visited_children: List[Node]) -> None:
        letter, _ = node.children
        self.key = letter.text

    def visit_quoted_value(self, node: Node, visited_children: List[Node]) -> None:
        self.value = unescape_string_value(node.text)

    def visit_number(self, node: Node, visited_children: List[Node]) -> None:
        self.value = node.text

    def generic_visit(self, node: Node, visited_children: List[Node]) -> None:
        ...


def is_command(line: str) -> bool:
    try:
        command_grammar.parse(line)
    except ParseError:
        return False
    else:
        return True


def parse_command(line: str) -> Tuple[str, Optional[str]]:
    try:
        return CommandVisitor().visit(command_grammar.parse(line))  # type: ignore
    except ParseError:
        if line.strip()[0] == "#":
            raise ParseError(
                "Line starts with '#' but it couldn't be parsed as a valid command"
            ) from None
        else:
            raise


CommandValue = Union[str, Path, int, Decimal, Fraction]


def dump_command(key: str, value: Optional[CommandValue] = None) -> str:
    if len(key) == 1:
        key_part = key
    else:
        key_part = f"#{key}"

    if value is None:
        value_part = ""
    else:
        dumped_value = dump_value(value)
        value_part = f"={dumped_value}"

    return key_part + value_part


@singledispatch
def dump_value(value: CommandValue) -> str:
    return str(value)


@dump_value.register
def dump_int_value(value: int) -> str:
    return str(value)


@dump_value.register
def dump_decimal_value(value: Decimal) -> str:
    return pretty_print_decimal(value)


@dump_value.register
def dump_fraction_value(value: Fraction) -> str:
    decimal = fraction_to_decimal(value)
    quantized = decimal.quantize(Decimal("0.000001"), rounding=ROUND_HALF_DOWN)
    return pretty_print_decimal(quantized)


@dump_value.register
def dump_string_value(value: str) -> str:
    """Escapes backslashes and " from a string"""
    escaped = escape_string_value(value)
    quoted = f'"{escaped}"'
    return quoted


@dump_value.register
def dump_path_value(value: Path) -> str:
    if value == Path(""):
        string_value = ""
    else:
        string_value = str(value)

    escaped = escape_string_value(string_value)
    quoted = f'"{escaped}"'
    return quoted


def pretty_print_decimal(d: Decimal) -> str:
    raw_string_form = str(d)
    if "." in raw_string_form:
        return raw_string_form.rstrip("0").rstrip(".")
    else:
        return raw_string_form


BACKSLASH = "\\"
ESCAPE_TABLE = str.maketrans({'"': BACKSLASH + '"', BACKSLASH: BACKSLASH + BACKSLASH})


def escape_string_value(unescaped: str) -> str:
    return unescaped.translate(ESCAPE_TABLE)


def unescape_string_value(escaped: str) -> str:
    """Unescapes a backslash-escaped string"""
    res = []
    i = 0
    while i < len(escaped):
        char = escaped[i]
        if char == BACKSLASH:
            if i + 1 == len(escaped):
                raise ValueError("backslash at end of string")
            else:
                i += 1

        res.append(escaped[i])
        i += 1

    return "".join(res)



"""
Hypothesis strategies to generate notes and charts
"""

from decimal import Decimal
from enum import Flag, auto
from functools import partial
from itertools import product
from pathlib import Path
from typing import Dict, Iterable, Optional, Set, Union

import hypothesis.strategies as st

from jubeatools.song import (
    BeatsTime,
    BPMEvent,
    Chart,
    Difficulty,
    LongNote,
    Metadata,
    NotePosition,
    Preview,
    Song,
    TapNote,
    Timing,
)


@st.composite
def beat_time(
    draw: st.DrawFn,
    min_section: Optional[int] = None,
    max_section: Optional[int] = None,
    min_numerator: Optional[int] = None,
    max_numerator: Optional[int] = None,
    denominator_strat: st.SearchStrategy[int] = st.sampled_from([4, 8, 16, 3, 5]),
) -> BeatsTime:
    denominator = draw(denominator_strat)

    if min_section is not None:
        min_value = denominator * 4 * min_section
    else:
        min_value = 0

    if min_numerator is not None:
        min_value = max(min_value, min_numerator)

    if max_section is not None:
        max_value: Optional[int] = denominator * 4 * max_section
    else:
        max_value = None

    if max_numerator is not None:
        if max_value is not None:
            max_value = min(max_numerator, max_value)
        else:
            max_value = max_numerator

    numerator = draw(st.integers(min_value=min_value, max_value=max_value))
    return BeatsTime(numerator, denominator)


@st.composite
def note_position(draw: st.DrawFn) -> NotePosition:
    x = draw(st.integers(min_value=0, max_value=3))
    y = draw(st.integers(min_value=0, max_value=3))
    return NotePosition(x, y)


@st.composite
def tap_note(
    draw: st.DrawFn,
    time_strat: st.SearchStrategy[BeatsTime] = beat_time(max_section=10),
) -> TapNote:
    time = draw(time_strat)
    position = draw(note_position())
    return TapNote(time, position)


@st.composite
def long_note(
    draw: st.DrawFn,
    time_strat: st.SearchStrategy[BeatsTime] = beat_time(max_section=10),
    duration_strat: st.SearchStrategy[BeatsTime] = beat_time(
        min_numerator=1, max_section=3
    ),
) -> LongNote:
    time = draw(time_strat)
    position = draw(note_position())
    duration = draw(duration_strat)
    tail_is_vertical = draw(st.booleans())
    tail_offset = draw(st.integers(min_value=1, max_value=3))
    if tail_is_vertical:
        x = position.x
        y = (position.y + tail_offset) % 4
    else:
        x = (position.x + tail_offset) % 4
        y = position.y
    tail_tip = NotePosition(x, y)
    return LongNote(time, position, duration, tail_tip)


@st.composite
def bad_notes(draw: st.DrawFn, longs: bool) -> Set[Union[TapNote, LongNote]]:
    note_strat = tap_note()
    if longs:
        note_strat = st.one_of(note_strat, long_note())
    notes: Set[Union[TapNote, LongNote]] = draw(st.sets(note_strat, max_size=32))
    return notes


@st.composite
def notes(
    draw: st.DrawFn,
    collisions: bool = False,
    note_strat: st.SearchStrategy[Union[TapNote, LongNote]] = st.one_of(
        tap_note(), long_note()
    ),
    beat_time_strat: st.SearchStrategy[BeatsTime] = beat_time(max_section=3),
) -> Set[Union[TapNote, LongNote]]:
    raw_notes: Set[Union[TapNote, LongNote]] = draw(st.sets(note_strat, max_size=32))

    if collisions:
        return raw_notes
    else:
        last_notes: Dict[NotePosition, Optional[BeatsTime]] = {
            NotePosition(x, y): None for y, x in product(range(4), range(4))
        }
        notes: Set[Union[TapNote, LongNote]] = set()
        for note in sorted(raw_notes, key=lambda n: (n.time, n.position)):
            last_note_time = last_notes[note.position]
            if last_note_time is None:
                new_time = draw(beat_time_strat)
            else:
                numerator = draw(
                    st.integers(min_value=1, max_value=last_note_time.denominator * 4)
                )
                distance = BeatsTime(numerator, last_note_time.denominator)
                new_time = last_note_time + distance
            if isinstance(note, LongNote):
                notes.add(
                    LongNote(
                        time=new_time,
                        position=note.position,
                        duration=note.duration,
                        tail_tip=note.tail_tip,
                    )
                )
                last_notes[note.position] = new_time + note.duration
            else:
                notes.add(TapNote(time=new_time, position=note.position))
                last_notes[note.position] = new_time
        return notes


@st.composite
def bpms(draw: st.DrawFn) -> Decimal:
    d: Decimal = draw(st.decimals(min_value=1, max_value=1000, places=3))
    return d


@st.composite
def bpm_changes(
    draw: st.DrawFn,
    bpm_strat: st.SearchStrategy[Decimal] = bpms(),
    time_strat: st.SearchStrategy[BeatsTime] = beat_time(min_section=1, max_section=10),
) -> BPMEvent:
    time = draw(time_strat)
    bpm = draw(bpm_strat)
    return BPMEvent(time, bpm)


@st.composite
def timing_info(
    draw: st.DrawFn,
    with_bpm_changes: bool = True,
    bpm_strat: st.SearchStrategy[Decimal] = bpms(),
    beat_zero_offset_strat: st.SearchStrategy[Decimal] = st.decimals(
        min_value=0, max_value=20, places=3
    ),
    time_strat: st.SearchStrategy[BeatsTime] = beat_time(min_section=1, max_section=10),
) -> Timing:
    first_bpm = draw(bpm_strat)
    first_event = BPMEvent(BeatsTime(0), first_bpm)
    events = [first_event]
    if with_bpm_changes:
        raw_bpm_changes = st.lists(
            bpm_changes(bpm_strat, time_strat), unique_by=get_bpm_change_time
        )
        sorted_bpm_changes = raw_bpm_changes.map(
            lambda l: sorted(l, key=get_bpm_change_time)
        )
        other_events = draw(sorted_bpm_changes)
        events += other_events
    beat_zero_offset = draw(beat_zero_offset_strat)
    return Timing(events=events, beat_zero_offset=beat_zero_offset)


def get_bpm_change_time(b: BPMEvent) -> BeatsTime:
    return b.time


@st.composite
def level(draw: st.DrawFn) -> Union[int, Decimal]:
    d: Union[int, Decimal] = draw(
        st.one_of(
            st.integers(min_value=0), st.decimals(min_value=0, max_value=10.9, places=1)
        )
    )
    return d


hakus = partial(st.sets, beat_time())


@st.composite
def chart(
    draw: st.DrawFn,
    timing_strat: st.SearchStrategy[Optional[Timing]] = timing_info(),
    hakus_strat: st.SearchStrategy[Optional[Set[BeatsTime]]] = st.none(),
    notes_strat: st.SearchStrategy[Iterable[Union[TapNote, LongNote]]] = notes(),
    level_strat: st.SearchStrategy[Union[int, Decimal]] = level(),
) -> Chart:
    level = Decimal(draw(level_strat))
    timing = draw(timing_strat)
    hakus = draw(hakus_strat)
    notes = draw(notes_strat)
    return Chart(
        level=level,
        timing=timing,
        hakus=hakus,
        notes=sorted(notes, key=lambda n: (n.time, n.position)),
    )


@st.composite
def preview(draw: st.DrawFn) -> Preview:
    start = draw(
        st.decimals(min_value=0, allow_nan=False, allow_infinity=False, places=3)
    )
    length = draw(
        st.decimals(min_value=1, allow_nan=False, allow_infinity=False, places=3)
    )
    return Preview(start, length)


metadata_text_strat = partial(
    st.text, alphabet=st.characters(blacklist_categories=("Cc", "Cs"))
)
metadata_path_strat = partial(
    st.text, alphabet=st.characters(blacklist_categories=("Cc", "Cs"))
)


@st.composite
def metadata(
    draw: st.DrawFn,
    text_strat: st.SearchStrategy[str] = metadata_text_strat(),
    path_strat: st.SearchStrategy[str] = metadata_path_strat(),
) -> Metadata:
    return Metadata(
        title=draw(text_strat),
        artist=draw(text_strat),
        audio=Path(draw(path_strat)),
        cover=Path(draw(path_strat)),
        preview=draw(st.one_of(st.none(), preview())),
        preview_file=Path(draw(path_strat)),
    )


class TimingOption(Flag):
    GLOBAL = auto()
    PER_CHART = auto()
    BPM_CHANGES = auto()


@st.composite
def song(
    draw: st.DrawFn,
    diffs_strat: st.SearchStrategy[Set[str]] = st.sets(
        st.sampled_from(list(d.value for d in Difficulty)), min_size=1, max_size=3
    ),
    common_timing_strat: st.SearchStrategy[Optional[Timing]] = timing_info(),
    common_hakus_strat: st.SearchStrategy[Optional[Set[BeatsTime]]] = st.none(),
    chart_strat: st.SearchStrategy[Chart] = chart(),
    metadata_strat: st.SearchStrategy[Metadata] = metadata(),
) -> Song:
    return Song(
        metadata=draw(metadata_strat),
        charts={difficulty: draw(chart_strat) for difficulty in draw(diffs_strat)},
        common_timing=draw(common_timing_strat),
        common_hakus=draw(common_hakus_strat),
    )



from decimal import Decimal
from pathlib import Path
from typing import List, Union

from hypothesis import example, given
from hypothesis import strategies as st

from jubeatools import song
from jubeatools.formats import Format
from jubeatools.formats.jubeat_analyser.memo1.dump import _dump_memo1_chart
from jubeatools.formats.jubeat_analyser.memo1.load import Memo1Parser
from jubeatools.testutils.strategies import notes as notes_strat
from jubeatools.testutils.test_patterns import dump_and_load_then_compare

from ..test_utils import memo_compatible_song
from . import example1


@given(notes_strat())
def test_that_notes_roundtrip(notes: List[Union[song.TapNote, song.LongNote]]) -> None:
    timing = song.Timing(
        events=[song.BPMEvent(song.BeatsTime(0), Decimal(120))],
        beat_zero_offset=song.SecondsTime(0),
    )
    chart = song.Chart(
        level=Decimal(0),
        timing=timing,
        notes=sorted(notes, key=lambda n: (n.time, n.position)),
    )
    metadata = song.Metadata("", "", Path(""), Path(""))
    string_io = _dump_memo1_chart("", chart, metadata, timing)
    chart_text = string_io.getvalue()
    parser = Memo1Parser()
    for line in chart_text.split("\n"):
        parser.load_line(line)
    parser.finish_last_few_notes()
    actual = set(parser.notes())
    assert notes == actual


@given(memo_compatible_song(), st.booleans())
@example(*example1.data)
def test_that_full_chart_roundtrips(song: song.Song, circle_free: bool) -> None:
    dump_and_load_then_compare(
        Format.MEMO_1,
        song,
        bytes_decoder=lambda b: b.decode("shift-jis-2004", errors="surrogateescape"),
        dump_options={"circle_free": circle_free},
    )



"""
Note symbol definition
"""
from decimal import Decimal
from typing import List, Optional, Tuple

from parsimonious import Grammar, NodeVisitor, ParseError
from parsimonious.nodes import Node

beat_symbol_line_grammar = Grammar(
    r"""
    line    = "*" symbol ":" number comment?
    symbol  = ws ~r"[^*#:|\-/\s]{1,2}" ws
    number  = ws ~r"\d+(\.\d+)?" ws
    ws      = ~r"[\t \u3000]*"
    comment = ~r"//.*"
    """
)


class BeatSymbolVisitor(NodeVisitor):
    def __init__(self) -> None:
        super().__init__()
        self.symbol: Optional[str] = None
        self.number: Optional[Decimal] = None

    def visit_line(
        self, node: Node, visited_children: List[Node]
    ) -> Tuple[str, Decimal]:
        if self.symbol is None:
            raise ValueError("No symbol found after parsing symbol definition")
        if self.number is None:
            raise ValueError("No value found after parsing symbol definition")
        return self.symbol, self.number

    def visit_symbol(self, node: Node, visited_children: List[Node]) -> None:
        _, symbol, _ = node.children
        self.symbol = symbol.text

    def visit_number(self, node: Node, visited_children: List[Node]) -> None:
        _, number, _ = node.children
        self.number = Decimal(number.text)

    def generic_visit(self, node: Node, visited_children: List[Node]) -> None:
        ...


def is_symbol_definition(line: str) -> bool:
    try:
        beat_symbol_line_grammar.parse(line)
    except ParseError:
        return False
    else:
        return True


def parse_symbol_definition(line: str) -> Tuple[str, Decimal]:
    return BeatSymbolVisitor().visit(beat_symbol_line_grammar.parse(line))  # type: ignore



from pathlib import Path
from typing import Any

from jubeatools import song
from jubeatools.formats.load_tools import make_folder_loader

from .. import commons as konami
from ..load_tools import make_chart_from_events
from . import construct


def load_jbsq(path: Path, *, beat_snap: int = 240, **kwargs: Any) -> song.Song:
    files = load_folder(path)
    charts = [
        load_jbsq_file(bytes_, path, beat_snap=beat_snap)
        for path, bytes_ in files.items()
    ]
    return song.Song.from_monochart_instances(*charts)


def load_file(path: Path) -> bytes:
    return path.read_bytes()


load_folder = make_folder_loader("*.jbsq", load_file)


def load_jbsq_file(
    bytes_: bytes, file_path: Path, *, beat_snap: int = 240
) -> song.Song:
    raw_data = construct.jbsq.parse(bytes_)
    events = [make_event_from_construct(e) for e in raw_data.events]
    chart = make_chart_from_events(events, beat_snap=beat_snap)
    dif = guess_difficulty(file_path.stem)
    return song.Song(metadata=song.Metadata(), charts={dif.value: chart})


def make_event_from_construct(e: construct.Event) -> konami.Event:
    return konami.Event(
        time=e.time_in_ticks,
        command=konami.Command[e.type_.name],
        value=e.value,
    )


def guess_difficulty(filename: str) -> song.Difficulty:
    try:
        return song.Difficulty(filename[-3:].upper())
    except ValueError:
        return song.Difficulty.EXTREME



"""
memo1

memo1 is an hybrid between youbeat and memo :

http://yosh52.web.fc2.com/jubeat/fumenformat.html

A chart in this format needs to have a `#memo1` line somewhere to indicate its format

It's very similar to memo, except it handles irregular timing bars which make
mono-column-style symbol definitions obsolete. Unlike #memo2 or youbeat however,
it does not handle in-bar bpm changes and pauses.
"""

from .dump import dump_memo1
from .load import load_memo1
